@@ -2,7 +2,7 @@
 
   gdbm.c -
 
-  $Author: akr $
+  $Author: nobu $
   modified at: Mon Jan 24 15:59:52 JST 1994
 
   Documentation by Peter Adolphs < futzilogik at users dot sourceforge dot net >
@@ -44,7 +44,7 @@
  * 1. Opening/creating a database, and filling it with some entries:
  *
  *      require 'gdbm'
- *      
+ *
  *      gdbm = GDBM.new("fruitstore.db")
  *      gdbm["ananas"]    = "3"
  *      gdbm["banana"]    = "8"
@@ -54,7 +54,7 @@
  * 2. Reading out a database:
  *
  *      require 'gdbm'
- *      
+ *
  *      gdbm = GDBM.new("fruitstore.db")
  *      gdbm.each_pair do |key, value|
  *        print "#{key}: #{value}\n"
@@ -73,12 +73,18 @@
  */
 static VALUE rb_cGDBM, rb_eGDBMError, rb_eGDBMFatalError;
 
+#if SIZEOF_LONG > SIZEOF_INT
+#define TOO_LONG(n) ((long)(+(int)(n)) != (long)(n))
+#else
+#define TOO_LONG(n) 0
+#endif
+
 #define RUBY_GDBM_RW_BIT 0x20000000
 
 #define MY_BLOCK_SIZE (2048)
 #define MY_FATAL_FUNC rb_gdbm_fatal
 static void
-rb_gdbm_fatal(char *msg)
+rb_gdbm_fatal(const char *msg)
 {
     rb_raise(rb_eGDBMFatalError, "%s", msg);
 }
@@ -95,25 +101,45 @@ closed_dbm(void)
 }
 
 #define GetDBM(obj, dbmp) do {\
-    Data_Get_Struct(obj, struct dbmdata, dbmp);\
-    if (dbmp == 0) closed_dbm();\
-    if (dbmp->di_dbm == 0) closed_dbm();\
+    TypedData_Get_Struct((obj), struct dbmdata, &dbm_type, (dbmp));\
+    if ((dbmp) == 0) closed_dbm();\
+    if ((dbmp)->di_dbm == 0) closed_dbm();\
 } while (0)
 
-#define GetDBM2(obj, data, dbm) {\
-    GetDBM(obj, data);\
-    (dbm) = dbmp->di_dbm;\
-}
+#define GetDBM2(obj, dbmp, dbm) do {\
+    GetDBM((obj), (dbmp));\
+    (dbm) = (dbmp)->di_dbm;\
+} while (0)
 
 static void
-free_dbm(struct dbmdata *dbmp)
+free_dbm(void *ptr)
 {
+    struct dbmdata *dbmp = ptr;
     if (dbmp) {
         if (dbmp->di_dbm) gdbm_close(dbmp->di_dbm);
-        free(dbmp);
+        xfree(dbmp);
     }
 }
 
+static size_t
+memsize_dbm(const void *ptr)
+{
+    size_t size = 0;
+    const struct dbmdata *dbmp = ptr;
+    if (dbmp) {
+	size += sizeof(*dbmp);
+	if (dbmp->di_dbm) size += DBM_SIZEOF_DBM;
+    }
+    return size;
+}
+
+static const rb_data_type_t dbm_type = {
+    "gdbm",
+    {0, free_dbm, memsize_dbm,},
+    0, 0,
+    RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
 /*
  * call-seq:
  *     gdbm.close -> nil
@@ -143,7 +169,7 @@ fgdbm_closed(VALUE obj)
 {
     struct dbmdata *dbmp;
 
-    Data_Get_Struct(obj, struct dbmdata, dbmp);
+    TypedData_Get_Struct(obj, struct dbmdata, &dbm_type, dbmp);
     if (dbmp == 0)
         return Qtrue;
     if (dbmp->di_dbm == 0)
@@ -155,7 +181,7 @@ fgdbm_closed(VALUE obj)
 static VALUE
 fgdbm_s_alloc(VALUE klass)
 {
-    return Data_Wrap_Struct(klass, 0, free_dbm, 0);
+    return TypedData_Wrap_Struct(klass, &dbm_type, 0);
 }
 
 /*
@@ -204,31 +230,40 @@ fgdbm_initialize(int argc, VALUE *argv, VALUE obj)
 
     SafeStringValue(file);
 
+#ifdef GDBM_CLOEXEC
+    /* GDBM_CLOEXEC is available since gdbm 1.10. */
+    flags |= GDBM_CLOEXEC;
+#endif
+
     if (flags & RUBY_GDBM_RW_BIT) {
         flags &= ~RUBY_GDBM_RW_BIT;
-        dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE, 
+        dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE,
                         flags, mode, MY_FATAL_FUNC);
     }
     else {
         dbm = 0;
         if (mode >= 0)
-            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE, 
+            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE,
                             GDBM_WRCREAT|flags, mode, MY_FATAL_FUNC);
         if (!dbm)
-            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE, 
+            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE,
                             GDBM_WRITER|flags, 0, MY_FATAL_FUNC);
         if (!dbm)
-            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE, 
+            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE,
                             GDBM_READER|flags, 0, MY_FATAL_FUNC);
     }
 
+    if (dbm) {
+        rb_fd_fix_cloexec(gdbm_fdesc(dbm));
+    }
+
     if (!dbm) {
         if (mode == -1) return Qnil;
 
         if (gdbm_errno == GDBM_FILE_OPEN_ERROR ||
             gdbm_errno == GDBM_CANT_BE_READER ||
             gdbm_errno == GDBM_CANT_BE_WRITER)
-            rb_sys_fail(RSTRING_PTR(file));
+            rb_sys_fail_str(file);
         else
             rb_raise(rb_eGDBMError, "%s", gdbm_strerror(gdbm_errno));
     }
@@ -264,7 +299,7 @@ fgdbm_initialize(int argc, VALUE *argv, VALUE obj)
 static VALUE
 fgdbm_s_open(int argc, VALUE *argv, VALUE klass)
 {
-    VALUE obj = Data_Wrap_Struct(klass, 0, free_dbm, 0);
+    VALUE obj = fgdbm_s_alloc(klass);
 
     if (NIL_P(fgdbm_initialize(argc, argv, obj))) {
         return Qnil;
@@ -290,17 +325,20 @@ rb_gdbm_fetch(GDBM_FILE dbm, datum key)
     str = rb_str_new(val.dptr, val.dsize);
     free(val.dptr);
     OBJ_TAINT(str);
-    return (VALUE)str;
+    return str;
 }
 
 static VALUE
 rb_gdbm_fetch2(GDBM_FILE dbm, VALUE keystr)
 {
     datum key;
+    long len;
 
     StringValue(keystr);
+    len = RSTRING_LEN(keystr);
+    if (TOO_LONG(len)) return Qnil;
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = (int)len;
 
     return rb_gdbm_fetch(dbm, key);
 }
@@ -336,14 +374,18 @@ rb_gdbm_nextkey(GDBM_FILE dbm, VALUE keystr)
 {
     datum key, key2;
     VALUE str;
+    long len;
 
+    len = RSTRING_LEN(keystr);
+    if (TOO_LONG(len)) return Qnil;
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = (int)len;
     key2 = gdbm_nextkey(dbm, key);
     if (key2.dptr == 0)
         return Qnil;
 
     str = rb_str_new(key2.dptr, key2.dsize);
+    free(key2.dptr);
     OBJ_TAINT(str);
     return str;
 }
@@ -396,13 +438,13 @@ fgdbm_fetch_m(int argc, VALUE *argv, VALUE obj)
 
 /*
  * call-seq:
- *      gdbm.index(value) -> key
+ *      gdbm.key(value) -> key
  *
  * Returns the _key_ for a given _value_. If several keys may map to the
  * same value, the key that is found first will be returned.
  */
 static VALUE
-fgdbm_index(VALUE obj, VALUE valstr)
+fgdbm_key(VALUE obj, VALUE valstr)
 {
     struct dbmdata *dbmp;
     GDBM_FILE dbm;
@@ -415,20 +457,28 @@ fgdbm_index(VALUE obj, VALUE valstr)
 
         valstr2 = rb_gdbm_fetch2(dbm, keystr);
         if (!NIL_P(valstr2) &&
-            RSTRING_LEN(valstr) == RSTRING_LEN(valstr2) &&
+            (int)RSTRING_LEN(valstr) == (int)RSTRING_LEN(valstr2) &&
             memcmp(RSTRING_PTR(valstr), RSTRING_PTR(valstr2),
-                   RSTRING_LEN(valstr)) == 0) {
+                   (int)RSTRING_LEN(valstr)) == 0) {
             return keystr;
         }
     }
     return Qnil;
 }
 
+/* :nodoc: */
+static VALUE
+fgdbm_index(VALUE obj, VALUE value)
+{
+    rb_warn("GDBM#index is deprecated; use GDBM#key");
+    return fgdbm_key(obj, value);
+}
+
 /*
  * call-seq:
- *      gdbm.select { |value| block } -> array
+ *      gdbm.select { |key, value| block } -> array
  *
- * Returns a new array of all values of the database for which _block_
+ * Returns a new array of all key-value pairs of the database for which _block_
  * evaluates to true.
  */
 static VALUE
@@ -476,7 +526,6 @@ fgdbm_values_at(int argc, VALUE *argv, VALUE obj)
 static void
 rb_gdbm_modify(VALUE obj)
 {
-    rb_secure(4);
     if (OBJ_FROZEN(obj)) rb_error_frozen("GDBM");
 }
 
@@ -486,11 +535,14 @@ rb_gdbm_delete(VALUE obj, VALUE keystr)
     datum key;
     struct dbmdata *dbmp;
     GDBM_FILE dbm;
+    long len;
 
     rb_gdbm_modify(obj);
     StringValue(keystr);
+    len = RSTRING_LEN(keystr);
+    if (TOO_LONG(len)) return Qnil;
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = (int)len;
 
     GetDBM2(obj, dbmp, dbm);
     if (!gdbm_exists(dbm, key)) {
@@ -528,7 +580,7 @@ fgdbm_delete(VALUE obj, VALUE keystr)
  * call-seq:
  *      gdbm.shift -> (key, value) or nil
  *
- * Removes a key-value-pair from this database and returns it as a 
+ * Removes a key-value-pair from this database and returns it as a
  * two-item array [ _key_, _value_ ]. Returns nil if the database is empty.
  */
 static VALUE
@@ -561,8 +613,9 @@ fgdbm_delete_if(VALUE obj)
     struct dbmdata *dbmp;
     GDBM_FILE dbm;
     VALUE keystr, valstr;
-    VALUE ret, ary = rb_ary_new();
-    int i, status = 0, n;
+    VALUE ret, ary = rb_ary_tmp_new(0);
+    long i;
+    int status = 0, n;
 
     rb_gdbm_modify(obj);
     GetDBM2(obj, dbmp, dbm);
@@ -572,17 +625,19 @@ fgdbm_delete_if(VALUE obj)
     for (keystr = rb_gdbm_firstkey(dbm); RTEST(keystr);
          keystr = rb_gdbm_nextkey(dbm, keystr)) {
 
+	OBJ_FREEZE(keystr);
         valstr = rb_gdbm_fetch2(dbm, keystr);
-        ret = rb_protect(rb_yield, rb_assoc_new(keystr, valstr), &status);
+        ret = rb_protect(rb_yield, rb_assoc_new(rb_str_dup(keystr), valstr), &status);
         if (status != 0) break;
         if (RTEST(ret)) rb_ary_push(ary, keystr);
         GetDBM2(obj, dbmp, dbm);
     }
 
     for (i = 0; i < RARRAY_LEN(ary); i++)
-        rb_gdbm_delete(obj, RARRAY_PTR(ary)[i]);
+        rb_gdbm_delete(obj, RARRAY_AREF(ary, i));
     if (status) rb_jump_tag(status);
-    if (n > 0) dbmp->di_size = n - RARRAY_LEN(ary);
+    if (n > 0) dbmp->di_size = n - (int)RARRAY_LEN(ary);
+    rb_ary_clear(ary);
 
     return obj;
 }
@@ -610,7 +665,7 @@ fgdbm_clear(VALUE obj)
             free(key.dptr);
             rb_raise(rb_eGDBMError, "%s", gdbm_strerror(gdbm_errno));
         }
-        free(key.dptr); 
+        free(key.dptr);
     }
 #else
     while (key = gdbm_firstkey(dbm), key.dptr) {
@@ -674,10 +729,10 @@ fgdbm_store(VALUE obj, VALUE keystr, VALUE valstr)
     StringValue(valstr);
 
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = RSTRING_LENINT(keystr);
 
     val.dptr = RSTRING_PTR(valstr);
-    val.dsize = RSTRING_LEN(valstr);
+    val.dsize = RSTRING_LENINT(valstr);
 
     GetDBM2(obj, dbmp, dbm);
     dbmp->di_size = -1;
@@ -690,13 +745,15 @@ fgdbm_store(VALUE obj, VALUE keystr, VALUE valstr)
 }
 
 static VALUE
-update_i(VALUE pair, VALUE dbm)
+update_i(RB_BLOCK_CALL_FUNC_ARGLIST(pair, dbm))
 {
+    const VALUE *ptr;
     Check_Type(pair, T_ARRAY);
     if (RARRAY_LEN(pair) < 2) {
         rb_raise(rb_eArgError, "pair must be [key, value]");
     }
-    fgdbm_store(dbm, RARRAY_PTR(pair)[0], RARRAY_PTR(pair)[1]);
+    ptr = RARRAY_CONST_PTR(pair);
+    fgdbm_store(dbm, ptr[0], ptr[1]);
     return Qnil;
 }
 
@@ -801,6 +858,8 @@ fgdbm_each_value(VALUE obj)
     GDBM_FILE dbm;
     VALUE keystr;
 
+    RETURN_ENUMERATOR(obj, 0, 0);
+
     GetDBM2(obj, dbmp, dbm);
     for (keystr = rb_gdbm_firstkey(dbm); RTEST(keystr);
          keystr = rb_gdbm_nextkey(dbm, keystr)) {
@@ -825,6 +884,8 @@ fgdbm_each_key(VALUE obj)
     GDBM_FILE dbm;
     VALUE keystr;
 
+    RETURN_ENUMERATOR(obj, 0, 0);
+
     GetDBM2(obj, dbmp, dbm);
     for (keystr = rb_gdbm_firstkey(dbm); RTEST(keystr);
          keystr = rb_gdbm_nextkey(dbm, keystr)) {
@@ -840,7 +901,7 @@ fgdbm_each_key(VALUE obj)
  *      gdbm.each_pair { |key, value| block } -> gdbm
  *
  * Executes _block_ for each key in the database, passing the _key_ and the
- * correspoding _value_ as a parameter.
+ * corresponding _value_ as a parameter.
  */
 static VALUE
 fgdbm_each_pair(VALUE obj)
@@ -849,6 +910,8 @@ fgdbm_each_pair(VALUE obj)
     struct dbmdata *dbmp;
     VALUE keystr;
 
+    RETURN_ENUMERATOR(obj, 0, 0);
+
     GetDBM2(obj, dbmp, dbm);
     for (keystr = rb_gdbm_firstkey(dbm); RTEST(keystr);
          keystr = rb_gdbm_nextkey(dbm, keystr)) {
@@ -912,7 +975,9 @@ fgdbm_values(VALUE obj)
 
 /*
  * call-seq:
+ *      gdbm.include?(k) -> true or false
  *      gdbm.has_key?(k) -> true or false
+ *      gdbm.member?(k) -> true or false
  *      gdbm.key?(k) -> true or false
  *
  * Returns true if the given key _k_ exists within the database.
@@ -924,10 +989,13 @@ fgdbm_has_key(VALUE obj, VALUE keystr)
     datum key;
     struct dbmdata *dbmp;
     GDBM_FILE dbm;
+    long len;
 
     StringValue(keystr);
+    len = RSTRING_LENINT(keystr);
+    if (TOO_LONG(len)) return Qfalse;
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = (int)len;
 
     GetDBM2(obj, dbmp, dbm);
     if (gdbm_exists(dbm, key))
@@ -958,9 +1026,9 @@ fgdbm_has_value(VALUE obj, VALUE valstr)
         valstr2 = rb_gdbm_fetch2(dbm, keystr);
 
         if (!NIL_P(valstr2) &&
-            RSTRING_LEN(valstr) == RSTRING_LEN(valstr2) &&
+            (int)RSTRING_LEN(valstr) == (int)RSTRING_LEN(valstr2) &&
             memcmp(RSTRING_PTR(valstr), RSTRING_PTR(valstr2),
-                   RSTRING_LEN(valstr)) == 0) {
+                   (int)RSTRING_LEN(valstr)) == 0) {
             return Qtrue;
         }
     }
@@ -1008,6 +1076,7 @@ fgdbm_reorganize(VALUE obj)
     rb_gdbm_modify(obj);
     GetDBM2(obj, dbmp, dbm);
     gdbm_reorganize(dbm);
+    rb_fd_fix_cloexec(gdbm_fdesc(dbm));
     return obj;
 }
 
@@ -1016,7 +1085,7 @@ fgdbm_reorganize(VALUE obj)
  *     gdbm.sync -> gdbm
  *
  * Unless the _gdbm_ object has been opened with the *SYNC* flag, it is not
- * guarenteed that database modification operations are immediately applied to
+ * guaranteed that database modification operations are immediately applied to
  * the database file. This method ensures that all recent modifications
  * to the database are written to the file. Blocks until all writing operations
  * to the disk have been finished.
@@ -1174,6 +1243,7 @@ Init_gdbm(void)
     rb_define_method(rb_cGDBM, "[]=", fgdbm_store, 2);
     rb_define_method(rb_cGDBM, "store", fgdbm_store, 2);
     rb_define_method(rb_cGDBM, "index",  fgdbm_index, 1);
+    rb_define_method(rb_cGDBM, "key",  fgdbm_key, 1);
     rb_define_method(rb_cGDBM, "select",  fgdbm_select, 0);
     rb_define_method(rb_cGDBM, "values_at",  fgdbm_values_at, -1);
     rb_define_method(rb_cGDBM, "length", fgdbm_length, 0);
