@@ -2,13 +2,16 @@
 
   math.c -
 
-  $Author: usa $
+  $Author: naruse $
   created at: Tue Jan 25 14:12:56 JST 1994
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
 
 **********************************************************************/
 
+#ifdef _MSC_VER
+# define _USE_MATH_DEFINES 1
+#endif
 #include "internal.h"
 #include <float.h>
 #include <math.h>
@@ -60,7 +63,7 @@ VALUE rb_eMathDomainError;
  */
 
 static VALUE
-math_atan2(VALUE obj, VALUE y, VALUE x)
+math_atan2(VALUE unused_obj, VALUE y, VALUE x)
 {
     double dx, dy;
     dx = Get_Double(x);
@@ -105,7 +108,7 @@ math_atan2(VALUE obj, VALUE y, VALUE x)
  */
 
 static VALUE
-math_cos(VALUE obj, VALUE x)
+math_cos(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(cos(Get_Double(x)));
 }
@@ -126,7 +129,7 @@ math_cos(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_sin(VALUE obj, VALUE x)
+math_sin(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(sin(Get_Double(x)));
 }
@@ -147,7 +150,7 @@ math_sin(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_tan(VALUE obj, VALUE x)
+math_tan(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(tan(Get_Double(x)));
 }
@@ -167,7 +170,7 @@ math_tan(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_acos(VALUE obj, VALUE x)
+math_acos(VALUE unused_obj, VALUE x)
 {
     double d;
 
@@ -191,7 +194,7 @@ math_acos(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_asin(VALUE obj, VALUE x)
+math_asin(VALUE unused_obj, VALUE x)
 {
     double d;
 
@@ -215,7 +218,7 @@ math_asin(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_atan(VALUE obj, VALUE x)
+math_atan(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(atan(Get_Double(x)));
 }
@@ -243,7 +246,7 @@ cosh(double x)
  */
 
 static VALUE
-math_cosh(VALUE obj, VALUE x)
+math_cosh(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(cosh(Get_Double(x)));
 }
@@ -271,7 +274,7 @@ sinh(double x)
  */
 
 static VALUE
-math_sinh(VALUE obj, VALUE x)
+math_sinh(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(sinh(Get_Double(x)));
 }
@@ -280,7 +283,14 @@ math_sinh(VALUE obj, VALUE x)
 double
 tanh(double x)
 {
-    return sinh(x) / cosh(x);
+# if defined(HAVE_SINH) && defined(HAVE_COSH)
+    const double c = cosh(x);
+    if (!isinf(c)) return sinh(x) / c;
+# else
+    const double e = exp(x+x);
+    if (!isinf(e)) return (e - 1) / (e + 1);
+# endif
+    return x > 0 ? 1.0 : -1.0;
 }
 #endif
 
@@ -299,7 +309,7 @@ tanh(double x)
  */
 
 static VALUE
-math_tanh(VALUE obj, VALUE x)
+math_tanh(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(tanh(Get_Double(x)));
 }
@@ -319,7 +329,7 @@ math_tanh(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_acosh(VALUE obj, VALUE x)
+math_acosh(VALUE unused_obj, VALUE x)
 {
     double d;
 
@@ -344,7 +354,7 @@ math_acosh(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_asinh(VALUE obj, VALUE x)
+math_asinh(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(asinh(Get_Double(x)));
 }
@@ -364,7 +374,7 @@ math_asinh(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_atanh(VALUE obj, VALUE x)
+math_atanh(VALUE unused_obj, VALUE x)
 {
     double d;
 
@@ -394,7 +404,7 @@ math_atanh(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_exp(VALUE obj, VALUE x)
+math_exp(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(exp(Get_Double(x)));
 }
@@ -408,6 +418,13 @@ math_exp(VALUE obj, VALUE x)
 # define log10(x) ((x) < 0.0 ? nan("") : log10(x))
 #endif
 
+#ifndef M_LN2
+# define M_LN2 0.693147180559945309417232121458176568
+#endif
+#ifndef M_LN10
+# define M_LN10 2.30258509299404568401799145468436421
+#endif
+
 static double math_log1(VALUE x);
 
 /*
@@ -432,7 +449,7 @@ static double math_log1(VALUE x);
  */
 
 static VALUE
-math_log(int argc, const VALUE *argv, VALUE obj)
+math_log(int argc, const VALUE *argv, VALUE unused_obj)
 {
     VALUE x, base;
     double d;
@@ -446,9 +463,8 @@ math_log(int argc, const VALUE *argv, VALUE obj)
 }
 
 static double
-math_log1(VALUE x)
+get_double_rshift(VALUE x, size_t *pnumbits)
 {
-    double d;
     size_t numbits;
 
     if (RB_BIGNUM_TYPE_P(x) && BIGNUM_POSITIVE_P(x) &&
@@ -459,14 +475,22 @@ math_log1(VALUE x)
     else {
 	numbits = 0;
     }
+    *pnumbits = numbits;
+    return Get_Double(x);
+}
+
+static double
+math_log1(VALUE x)
+{
+    size_t numbits;
+    double d = get_double_rshift(x, &numbits);
 
-    d = Get_Double(x);
     /* check for domain error */
     if (d < 0.0) domain_error("log");
     /* check for pole error */
     if (d == 0.0) return -INFINITY;
 
-    return log(d) + numbits * log(2); /* log(d * 2 ** numbits) */
+    return log(d) + numbits * M_LN2; /* log(d * 2 ** numbits) */
 }
 
 #ifndef log2
@@ -499,21 +523,11 @@ extern double log2(double);
  */
 
 static VALUE
-math_log2(VALUE obj, VALUE x)
+math_log2(VALUE unused_obj, VALUE x)
 {
-    double d;
     size_t numbits;
+    double d = get_double_rshift(x, &numbits);
 
-    if (RB_BIGNUM_TYPE_P(x) && BIGNUM_POSITIVE_P(x) &&
-            DBL_MAX_EXP <= (numbits = rb_absint_numwords(x, 1, NULL))) {
-        numbits -= DBL_MANT_DIG;
-        x = rb_big_rshift(x, SIZET2NUM(numbits));
-    }
-    else {
-	numbits = 0;
-    }
-
-    d = Get_Double(x);
     /* check for domain error */
     if (d < 0.0) domain_error("log2");
     /* check for pole error */
@@ -539,21 +553,11 @@ math_log2(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_log10(VALUE obj, VALUE x)
+math_log10(VALUE unused_obj, VALUE x)
 {
-    double d;
     size_t numbits;
+    double d = get_double_rshift(x, &numbits);
 
-    if (RB_BIGNUM_TYPE_P(x) && BIGNUM_POSITIVE_P(x) &&
-            DBL_MAX_EXP <= (numbits = rb_absint_numwords(x, 1, NULL))) {
-        numbits -= DBL_MANT_DIG;
-        x = rb_big_rshift(x, SIZET2NUM(numbits));
-    }
-    else {
-	numbits = 0;
-    }
-
-    d = Get_Double(x);
     /* check for domain error */
     if (d < 0.0) domain_error("log10");
     /* check for pole error */
@@ -589,10 +593,43 @@ math_log10(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_sqrt(VALUE obj, VALUE x)
+math_sqrt(VALUE unused_obj, VALUE x)
+{
+    return rb_math_sqrt(x);
+}
+
+#define f_boolcast(x) ((x) ? Qtrue : Qfalse)
+inline static VALUE
+f_negative_p(VALUE x)
+{
+    if (FIXNUM_P(x))
+        return f_boolcast(FIX2LONG(x) < 0);
+    return rb_funcall(x, '<', 1, INT2FIX(0));
+}
+inline static VALUE
+f_signbit(VALUE x)
+{
+    if (RB_TYPE_P(x, T_FLOAT)) {
+        double f = RFLOAT_VALUE(x);
+        return f_boolcast(!isnan(f) && signbit(f));
+    }
+    return f_negative_p(x);
+}
+
+VALUE
+rb_math_sqrt(VALUE x)
 {
     double d;
 
+    if (RB_TYPE_P(x, T_COMPLEX)) {
+	VALUE neg = f_signbit(RCOMPLEX(x)->imag);
+	double re = Get_Double(RCOMPLEX(x)->real), im;
+	d = Get_Double(rb_complex_abs(x));
+	im = sqrt((d - re) / 2.0);
+	re = sqrt((d + re) / 2.0);
+	if (neg) im = -im;
+	return rb_complex_new(DBL2NUM(re), DBL2NUM(im));
+    }
     d = Get_Double(x);
     /* check for domain error */
     if (d < 0.0) domain_error("sqrt");
@@ -636,7 +673,7 @@ math_sqrt(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_cbrt(VALUE obj, VALUE x)
+math_cbrt(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(cbrt(Get_Double(x)));
 }
@@ -646,14 +683,14 @@ math_cbrt(VALUE obj, VALUE x)
  *     Math.frexp(x)    -> [fraction, exponent]
  *
  *  Returns a two-element array containing the normalized fraction (a Float)
- *  and exponent (a Fixnum) of +x+.
+ *  and exponent (an Integer) of +x+.
  *
  *     fraction, exponent = Math.frexp(1234)   #=> [0.6025390625, 11]
  *     fraction * 2**exponent                  #=> 1234.0
  */
 
 static VALUE
-math_frexp(VALUE obj, VALUE x)
+math_frexp(VALUE unused_obj, VALUE x)
 {
     double d;
     int exp;
@@ -673,7 +710,7 @@ math_frexp(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_ldexp(VALUE obj, VALUE x, VALUE n)
+math_ldexp(VALUE unused_obj, VALUE x, VALUE n)
 {
     return DBL2NUM(ldexp(Get_Double(x), NUM2INT(n)));
 }
@@ -689,7 +726,7 @@ math_ldexp(VALUE obj, VALUE x, VALUE n)
  */
 
 static VALUE
-math_hypot(VALUE obj, VALUE x, VALUE y)
+math_hypot(VALUE unused_obj, VALUE x, VALUE y)
 {
     return DBL2NUM(hypot(Get_Double(x), Get_Double(y)));
 }
@@ -709,7 +746,7 @@ math_hypot(VALUE obj, VALUE x, VALUE y)
  */
 
 static VALUE
-math_erf(VALUE obj, VALUE x)
+math_erf(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(erf(Get_Double(x)));
 }
@@ -729,7 +766,7 @@ math_erf(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_erfc(VALUE obj, VALUE x)
+math_erfc(VALUE unused_obj, VALUE x)
 {
     return DBL2NUM(erfc(Get_Double(x)));
 }
@@ -810,7 +847,7 @@ ruby_lgamma_r(const double d, int *sign)
  */
 
 static VALUE
-math_gamma(VALUE obj, VALUE x)
+math_gamma(VALUE unused_obj, VALUE x)
 {
     static const double fact_table[] = {
         /* fact(0) */ 1.0,
@@ -869,7 +906,7 @@ math_gamma(VALUE obj, VALUE x)
  */
 
 static VALUE
-math_lgamma(VALUE obj, VALUE x)
+math_lgamma(VALUE unused_obj, VALUE x)
 {
     double d;
     int sign=1;
@@ -889,14 +926,14 @@ math_lgamma(VALUE obj, VALUE x)
 VALUE \
 rb_math_##n(VALUE x)\
 {\
-    return math_##n(rb_mMath, x);\
+    return math_##n(0, x);\
 }
 
 #define exp2(n) \
 VALUE \
 rb_math_##n(VALUE x, VALUE y)\
 {\
-    return math_##n(rb_mMath, x, y);\
+    return math_##n(0, x, y);\
 }
 
 exp2(atan2)
@@ -908,7 +945,7 @@ exp2(hypot)
 VALUE
 rb_math_log(int argc, const VALUE *argv)
 {
-    return math_log(argc, argv, rb_mMath);
+    return math_log(argc, argv, 0);
 }
 
 exp1(sin)
@@ -952,12 +989,8 @@ InitVM_Math(void)
     rb_mMath = rb_define_module("Math");
     rb_eMathDomainError = rb_define_class_under(rb_mMath, "DomainError", rb_eStandardError);
 
-#ifdef M_PI
     /*  Definition of the mathematical constant PI as a Float number. */
     rb_define_const(rb_mMath, "PI", DBL2NUM(M_PI));
-#else
-    rb_define_const(rb_mMath, "PI", DBL2NUM(atan(1.0)*4.0));
-#endif
 
 #ifdef M_E
     /*  Definition of the mathematical constant E (e) as a Float number. */
