@@ -261,10 +261,15 @@ rb_gcd_gmp(VALUE x, VALUE y)
 
     mpz_gcd(mz, mx, my);
 
+    mpz_clear(mx);
+    mpz_clear(my);
+
     zn = (mpz_sizeinbase(mz, 16) + SIZEOF_BDIGIT*2 - 1) / (SIZEOF_BDIGIT*2);
     z = rb_big_new(zn, 1);
     mpz_export(BIGNUM_DIGITS(z), &count, -1, sizeof(BDIGIT), 0, nails, mz);
 
+    mpz_clear(mz);
+
     return rb_big_norm(z);
 }
 #endif
@@ -721,7 +726,7 @@ static double nurat_to_double(VALUE self);
  * Performs addition.
  *
  *    Rational(2, 3)  + Rational(2, 3)   #=> (4/3)
- *    Rational(900)   + Rational(1)      #=> (900/1)
+ *    Rational(900)   + Rational(1)      #=> (901/1)
  *    Rational(-2, 9) + Rational(-9, 2)  #=> (-85/18)
  *    Rational(9, 8)  + 4                #=> (41/8)
  *    Rational(20, 9) + 9.8              #=> 12.022222222222222
@@ -1034,6 +1039,14 @@ nurat_expt(VALUE self, VALUE other)
 		num = ONE;
 		den = ONE;
 	    }
+	    if (RB_FLOAT_TYPE_P(num)) { /* infinity due to overflow */
+		if (RB_FLOAT_TYPE_P(den)) return DBL2NUM(NAN);
+		return num;
+	    }
+	    if (RB_FLOAT_TYPE_P(den)) { /* infinity due to overflow */
+		num = ZERO;
+		den = ONE;
+	    }
 	    return f_rational_new2(CLASS_OF(self), num, den);
 	}
     }
@@ -2365,11 +2378,22 @@ read_num(const char **s, int numsign, int strict,
 	    exp = rb_int_uminus(exp);
     }
 
-    if (numsign == '-')
-	*num = rb_rational_uminus(*num);
+    if (numsign == '-') {
+	if (RB_TYPE_P(*num, T_RATIONAL)) {
+	    *num = rb_rational_uminus(*num);
+	}
+	else {
+	    *num = rb_int_uminus(*num);
+	}
+    }
     if (!NIL_P(exp)) {
 	VALUE l = f_expt10(exp);
-	*num = nurat_mul(*num, l);
+	if (RB_TYPE_P(*num, T_RATIONAL)) {
+	    *num = nurat_mul(*num, l);
+	}
+	else {
+	    *num = rb_int_mul(*num, l);
+	}
     }
     return 1;
 }
@@ -2395,8 +2419,14 @@ read_rat_nos(const char **s, int sign, int strict,
 	(*s)++;
 	if (!read_den(s, strict, &den))
 	    return 0;
-	if (!(FIXNUM_P(den) && FIX2LONG(den) == 1))
-	    *num = nurat_div(*num, den);
+	if (!(FIXNUM_P(den) && FIX2LONG(den) == 1)) {
+	    if (RB_TYPE_P(*num, T_RATIONAL)) {
+		*num = nurat_div(*num, den);
+	    }
+	    else {
+		*num = rb_int_div(*num, den);
+	    }
+	}
     }
     return 1;
 }
