@@ -48,7 +48,9 @@
  */
 
 #include "ruby/ruby.h"
+#include "ruby/encoding.h"
 #include "timev.h"
+#include "internal.h"
 
 #ifndef GAWK
 #include <stdio.h>
@@ -73,9 +75,6 @@
 #define VMS_EXT		1	/* include %v for VMS date format */
 #define MAILHEADER_EXT	1	/* add %z for HHMM format */
 #define ISO_DATE_EXT	1	/* %G and %g for year of ISO week */
-#ifndef GAWK
-#define POSIX_SEMANTICS	1	/* call tzset() if TZ changes */
-#endif
 
 #if defined(ISO_DATE_EXT)
 #if ! defined(POSIX2_DATE)
@@ -123,39 +122,14 @@ extern char *getenv();
 extern char *strchr();
 #endif
 
-#define range(low, item, hi)	max(low, min(item, hi))
-
-#if defined __WIN32__ || defined _WIN32
-#define DLL_IMPORT __declspec(dllimport)
-#endif
-#ifndef DLL_IMPORT
-#define DLL_IMPORT
-#endif
-#if !defined(OS2) && defined(HAVE_TZNAME)
-extern DLL_IMPORT char *tzname[2];
-#ifdef HAVE_DAYLIGHT
-extern DLL_IMPORT int daylight;
-#endif
-#ifdef HAVE_VAR_TIMEZONE
-extern DLL_IMPORT TYPEOF_VAR_TIMEZONE timezone;
-#endif
-#ifdef HAVE_VAR_ALTZONE
-extern DLL_IMPORT TYPEOF_VAR_ALTZONE altzone;
-#endif
-#endif
+#define range(low, item, hi)	max((low), min((item), (hi)))
 
 #undef min	/* just in case */
 
 /* min --- return minimum of two numbers */
 
-#ifndef __STDC__
-static inline int
-min(a, b)
-int a, b;
-#else
 static inline int
 min(int a, int b)
-#endif
 {
 	return (a < b ? a : b);
 }
@@ -164,14 +138,8 @@ min(int a, int b)
 
 /* max --- return maximum of two numbers */
 
-#ifndef __STDC__
-static inline int
-max(a, b)
-int a, b;
-#else
 static inline int
 max(int a, int b)
-#endif
 {
 	return (a > b ? a : b);
 }
@@ -189,36 +157,95 @@ max(int a, int b)
 
 /* strftime --- produce formatted time */
 
-static size_t
-rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const struct vtm *vtm, VALUE timev, struct timespec *ts, int gmt)
+enum {LEFT, CHCASE, LOWER, UPPER};
+#define BIT_OF(n) (1U<<(n))
+
+static char *
+resize_buffer(VALUE ftime, char *s, const char **start, const char **endp,
+	      ptrdiff_t n, size_t maxsize)
 {
-	char *endp = s + maxsize;
-	char *start = s;
+	size_t len = s - *start;
+	size_t nlen = len + n * 2;
+
+	if (nlen < len || nlen > maxsize) {
+		return 0;
+	}
+	rb_str_set_len(ftime, len);
+	rb_str_modify_expand(ftime, nlen-len);
+	s = RSTRING_PTR(ftime);
+	*endp = s + nlen;
+	*start = s;
+	return s += len;
+}
+
+static void
+buffer_size_check(const char *s,
+		  const char *format_end, size_t format_len,
+		  rb_encoding *enc)
+{
+	if (!s) {
+		const char *format = format_end-format_len;
+		VALUE fmt = rb_enc_str_new(format, format_len, enc);
+		rb_syserr_fail_str(ERANGE, fmt);
+	}
+}
+
+static char *
+case_conv(char *s, ptrdiff_t i, int flags)
+{
+	switch (flags & (BIT_OF(UPPER)|BIT_OF(LOWER))) {
+	case BIT_OF(UPPER):
+		do {
+			if (ISLOWER(*s)) *s = TOUPPER(*s);
+		} while (s++, --i);
+		break;
+	case BIT_OF(LOWER):
+		do {
+			if (ISUPPER(*s)) *s = TOLOWER(*s);
+		} while (s++, --i);
+		break;
+	default:
+		s += i;
+		break;
+	}
+	return s;
+}
+
+static VALUE
+format_value(VALUE val, int base)
+{
+	if (!RB_TYPE_P(val, T_BIGNUM))
+		val = rb_Integer(val);
+	return rb_big2str(val, base);
+}
+
+/*
+ * enc is the encoding of the format. It is used as the encoding of resulted
+ * string, but the name of the month and weekday are always US-ASCII. So it
+ * is only used for the timezone name on Windows.
+ */
+static VALUE
+rb_strftime_with_timespec(VALUE ftime, const char *format, size_t format_len,
+			  rb_encoding *enc, const struct vtm *vtm, VALUE timev,
+			  struct timespec *ts, int gmt, size_t maxsize)
+{
+	size_t len = RSTRING_LEN(ftime);
+	char *s = RSTRING_PTR(ftime);
+	const char *start = s;
+	const char *endp = start + rb_str_capacity(ftime);
+	const char *const format_end = format + format_len;
 	const char *sp, *tp;
-	auto char tbuf[100];
+#define TBUFSIZE 100
+	auto char tbuf[TBUFSIZE];
 	long off;
 	ptrdiff_t i;
 	int w;
 	long y;
-	static short first = 1;
-#ifdef POSIX_SEMANTICS
-	static char *savetz = NULL;
-	static size_t savetzlen = 0;
-	char *tz;
-#endif /* POSIX_SEMANTICS */
-#ifndef HAVE_TM_ZONE
-#ifndef HAVE_TM_NAME
-#if ((defined(MAILHEADER_EXT) && !HAVE_VAR_TIMEZONE && HAVE_GETTIMEOFDAY) || \
-     (!HAVE_TZNAME && HAVE_TIMEZONE))
-	struct timeval tv;
-	struct timezone zone;
-#endif
-#endif /* HAVE_TM_NAME */
-#endif /* HAVE_TM_ZONE */
-	int precision, flags;
+	int precision, flags, colons;
 	char padding;
-	enum {LEFT, CHCASE, LOWER, UPPER, LOCALE_O, LOCALE_E};
-#define BIT_OF(n) (1U<<(n))
+#ifdef MAILHEADER_EXT
+	int sign;
+#endif
 
 	/* various tables, useful in North America */
 	static const char days_l[][10] = {
@@ -232,88 +259,77 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
 	};
 	static const char ampm[][3] = { "AM", "PM", };
 
-	if (s == NULL || format == NULL || vtm == NULL || maxsize == 0)
-		return 0;
-
-	/* quick check if we even need to bother */
-	if (strchr(format, '%') == NULL && strlen(format) + 1 >= maxsize) {
+	if (format == NULL || format_len == 0 || vtm == NULL) {
 	err:
-		errno = ERANGE;
 		return 0;
 	}
 
-#ifndef POSIX_SEMANTICS
-	if (first) {
-		tzset();
-		first = 0;
+	if (enc &&
+	    (enc == rb_usascii_encoding() ||
+	     enc == rb_ascii8bit_encoding() ||
+	     enc == rb_locale_encoding())) {
+		enc = NULL;
 	}
-#else	/* POSIX_SEMANTICS */
-	tz = getenv("TZ");
-	if (first) {
-		if (tz != NULL) {
-			size_t tzlen = strlen(tz);
 
-			savetz = (char *) malloc(tzlen + 1);
-			if (savetz != NULL) {
-				savetzlen = tzlen + 1;
-				memcpy(savetz, tz, savetzlen);
-			}
-		}
-		tzset();
-		first = 0;
-	}
-	/* if we have a saved TZ, and it is different, recapture and reset */
-	if (tz && savetz && (tz[0] != savetz[0] || strcmp(tz, savetz) != 0)) {
-		size_t i = strlen(tz) + 1;
-		if (i > savetzlen) {
-			savetz = (char *) realloc(savetz, i);
-			if (savetz) {
-				savetzlen = i;
-				memcpy(savetz, tz, i);
-			}
-		} else
-			memcpy(savetz, tz, i);
-		tzset();
-	}
-#endif	/* POSIX_SEMANTICS */
-
-	for (; *format && s < endp - 1; format++) {
+	s += len;
+	for (; format < format_end; format++) {
 #define FLAG_FOUND() do { \
-			if (precision > 0 || flags & (BIT_OF(LOCALE_E)|BIT_OF(LOCALE_O))) \
+			if (precision > 0) \
 				goto unknown; \
 		} while (0)
-#define NEEDS(n) do if (s + (n) >= endp - 1) goto err; while (0)
+#define NEEDS(n) do { \
+			if (s >= endp || (n) >= endp - s - 1) { \
+				s = resize_buffer(ftime, s, &start, &endp, (n), maxsize); \
+				buffer_size_check(s, format_end, format_len, enc); \
+			} \
+		} while (0)
 #define FILL_PADDING(i) do { \
-	if (!(flags & BIT_OF(LEFT)) && precision > i) { \
+	if (!(flags & BIT_OF(LEFT)) && precision > (i)) { \
 		NEEDS(precision); \
-		memset(s, padding ? padding : ' ', precision - i); \
-		s += precision - i; \
+		memset(s, padding ? padding : ' ', precision - (i)); \
+		s += precision - (i); \
 	} \
 	else { \
 		NEEDS(i); \
 	} \
 } while (0);
+#define FMT_PADDING(fmt, def_pad) \
+		(&"%*"fmt"\0""%0*"fmt[\
+			(padding == '0' || (!padding && (def_pad) == '0')) ? \
+			rb_strlen_lit("%*"fmt)+1 : 0])
+#define FMT_PRECISION(def_prec) \
+		((flags & BIT_OF(LEFT)) ? (1) : \
+		 (precision <= 0) ? (def_prec) : (precision))
 #define FMT(def_pad, def_prec, fmt, val) \
 		do { \
-			int l; \
-			if (precision <= 0) precision = (def_prec); \
-			if (flags & BIT_OF(LEFT)) precision = 1; \
-			l = snprintf(s, endp - s, \
-				     ((padding == '0' || (!padding && def_pad == '0')) ? "%0*"fmt : "%*"fmt), \
-				     precision, val); \
-			if (l < 0) goto err; \
-			s += l; \
+			precision = FMT_PRECISION(def_prec); \
+			len = s - start; \
+			NEEDS(precision); \
+			rb_str_set_len(ftime, len); \
+			rb_str_catf(ftime, FMT_PADDING(fmt, def_pad), \
+				    precision, (val)); \
+			RSTRING_GETMEM(ftime, s, len); \
+			endp = (start = s) + rb_str_capacity(ftime); \
+			s += len; \
 		} while (0)
 #define STRFTIME(fmt) \
 		do { \
-			i = rb_strftime_with_timespec(s, endp - s, fmt, vtm, timev, ts, gmt); \
-			if (!i) return 0; \
+			len = s - start; \
+			rb_str_set_len(ftime, len); \
+			if (!rb_strftime_with_timespec(ftime, (fmt), rb_strlen_lit(fmt), \
+						       enc, vtm, timev, ts, gmt, maxsize)) \
+				return 0; \
+			s = RSTRING_PTR(ftime); \
+			i = RSTRING_LEN(ftime) - len; \
+			endp = (start = s) + rb_str_capacity(ftime); \
+			s += len; \
+			if (i > 0) case_conv(s, i, flags); \
 			if (precision > i) {\
 				NEEDS(precision); \
 				memmove(s + precision - i, s, i);\
 				memset(s, padding ? padding : ' ', precision - i); \
 				s += precision;	\
-	                }\
+			} \
 			else s += i; \
 		} while (0)
 #define FMTV(def_pad, def_prec, fmt, val) \
@@ -323,38 +339,39 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
                                 FMT((def_pad), (def_prec), "l"fmt, FIX2LONG(tmp)); \
                         } \
                         else { \
-                                VALUE args[2], result; \
-                                size_t l; \
-                                if (precision <= 0) precision = (def_prec); \
-                                if (flags & BIT_OF(LEFT)) precision = 1; \
-                                args[0] = INT2FIX(precision); \
-                                args[1] = val; \
-                                if (padding == '0' || (!padding && def_pad == '0')) \
-                                        result = rb_str_format(2, args, rb_str_new2("%0*"fmt)); \
-                                else \
-                                        result = rb_str_format(2, args, rb_str_new2("%*"fmt)); \
-                                l = strlcpy(s, StringValueCStr(result), endp-s); \
-                                if ((size_t)(endp-s) <= l) \
-                                        goto err; \
-                                s += l; \
+				const int base = ((fmt[0] == 'x') ? 16 : \
+						  (fmt[0] == 'o') ? 8 : \
+						  10); \
+				precision = FMT_PRECISION(def_prec); \
+				if (!padding) padding = (def_pad); \
+				tmp = format_value(tmp, base); \
+				i = RSTRING_LEN(tmp); \
+				FILL_PADDING(i); \
+				rb_str_set_len(ftime, s-start); \
+				rb_str_append(ftime, tmp); \
+				RSTRING_GETMEM(ftime, s, len); \
+				endp = (start = s) + rb_str_capacity(ftime); \
+				s += len; \
                         } \
                 } while (0)
 
-		if (*format != '%') {
-			*s++ = *format;
-			continue;
-		}
+		tp = memchr(format, '%', format_end - format);
+		if (!tp) tp = format_end;
+		NEEDS(tp - format);
+		memcpy(s, format, tp - format);
+		s += tp - format;
+		format = tp;
+		if (format == format_end) break;
+
 		tp = tbuf;
 		sp = format;
 		precision = -1;
 		flags = 0;
 		padding = 0;
+                colons = 0;
 	again:
-		switch (*++format) {
-		case '\0':
-			format--;
-			goto unknown;
-
+		if (++format >= format_end) goto unknown;
+		switch (*format) {
 		case '%':
 			FILL_PADDING(1);
 			*s++ = '%';
@@ -431,7 +448,8 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
 			continue;
 
 		case 'j':	/* day of the year, 001 - 366 */
-			FMT('0', 3, "d", vtm->yday);
+			i = range(1, vtm->yday, 366);
+			FMT('0', 3, "d", (int)i);
 			continue;
 
 		case 'm':	/* month, 01 - 12 */
@@ -506,89 +524,87 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
 
 		case 'Y':	/* year with century */
                         if (FIXNUM_P(vtm->year)) {
-                            long y = FIX2LONG(vtm->year);
-                            FMT('0', 0 <= y ? 4 : 5, "ld", y);
+				long y = FIX2LONG(vtm->year);
+				FMT('0', 0 <= y ? 4 : 5, "ld", y);
                         }
                         else {
-                            FMTV('0', 4, "d", vtm->year);
+				FMTV('0', 4, "d", vtm->year);
                         }
 			continue;
 
 #ifdef MAILHEADER_EXT
-		/*
-		 * From: Chip Rosenthal <chip@chinacat.unicom.com>
-		 * Date: Sun, 19 Mar 1995 00:33:29 -0600 (CST)
-		 *
-		 * Warning: the %z [code] is implemented by inspecting the
-		 * timezone name conditional compile settings, and
-		 * inferring a method to get timezone offsets. I've tried
-		 * this code on a couple of machines, but I don't doubt
-		 * there is some system out there that won't like it.
-		 * Maybe the easiest thing to do would be to bracket this
-		 * with an #ifdef that can turn it off. The %z feature
-		 * would be an admittedly obscure one that most folks can
-		 * live without, but it would be a great help to those of
-		 * us that muck around with various message processors.
-		 */
 		case 'z':	/* time zone offset east of GMT e.g. -0600 */
-			if (precision < 4) precision = 4;
-			NEEDS(precision + 1);
 			if (gmt) {
 				off = 0;
 			}
 			else {
-				off = NUM2LONG(rb_funcall(quo(vtm->utc_offset, INT2FIX(60)), rb_intern("round"), 0));
-#if 0
-#ifdef HAVE_TM_NAME
-				/*
-				 * Systems with tm_name probably have tm_tzadj as
-				 * secs west of GMT.  Convert to mins east of GMT.
-				 */
-				off = -timeptr->tm_tzadj / 60;
-#else /* !HAVE_TM_NAME */
-#ifdef HAVE_TM_ZONE
-				/*
-				 * Systems with tm_zone probably have tm_gmtoff as
-				 * secs east of GMT.  Convert to mins east of GMT.
-				 */
-				off = timeptr->tm_gmtoff / 60;
-#else /* !HAVE_TM_ZONE */
-#if HAVE_VAR_TIMEZONE
-#if HAVE_VAR_ALTZONE
-				off = -(daylight ? timezone : altzone) / 60;
-#else
-				off = -timezone / 60;
-#endif
-#else /* !HAVE_VAR_TIMEZONE */
-#ifdef HAVE_GETTIMEOFDAY
-				gettimeofday(&tv, &zone);
-				off = -zone.tz_minuteswest;
-#else
-				/* no timezone info, then calc by myself */
-				{
-					struct tm utc;
-					time_t now;
-					time(&now);
-					utc = *gmtime(&now);
-					off = (long)((now - mktime(&utc)) / 60);
-				}
-#endif
-#endif /* !HAVE_VAR_TIMEZONE */
-#endif /* !HAVE_TM_ZONE */
-#endif /* !HAVE_TM_NAME */
-#endif /* 0 */
+				off = NUM2LONG(rb_funcall(vtm->utc_offset, rb_intern("round"), 0));
 			}
 			if (off < 0) {
 				off = -off;
-				*s++ = '-';
+				sign = -1;
 			} else {
-				*s++ = '+';
+				sign = +1;
 			}
-			off = off/60*100 + off%60;
-			i = snprintf(s, endp - s, (padding == ' ' ? "%*ld" : "%.*ld"),
-				     precision - (precision > 4), off);
+                        switch (colons) {
+			case 0: /* %z -> +hhmm */
+				precision = precision <= 5 ? 2 : precision-3;
+				NEEDS(precision + 3);
+				break;
+
+			case 1: /* %:z -> +hh:mm */
+				precision = precision <= 6 ? 2 : precision-4;
+				NEEDS(precision + 4);
+				break;
+
+			case 2: /* %::z -> +hh:mm:ss */
+				precision = precision <= 9 ? 2 : precision-7;
+				NEEDS(precision + 7);
+				break;
+
+			case 3: /* %:::z -> +hh[:mm[:ss]] */
+				if (off % 3600 == 0) {
+					precision = precision <= 3 ? 2 : precision-1;
+					NEEDS(precision + 3);
+				}
+				else if (off % 60 == 0) {
+					precision = precision <= 6 ? 2 : precision-4;
+					NEEDS(precision + 4);
+				}
+				else {
+					precision = precision <= 9 ? 2 : precision-7;
+					NEEDS(precision + 9);
+				}
+				break;
+
+			default:
+				format--;
+				goto unknown;
+                        }
+			i = snprintf(s, endp - s, (padding == ' ' ? "%+*ld" : "%+.*ld"),
+				     precision + 1, sign * (off / 3600));
+			if (i < 0) goto err;
+			if (sign < 0 && off < 3600) {
+				*(padding == ' ' ? s + i - 2 : s) = '-';
+			}
+			s += i;
+                        off = off % 3600;
+			if (colons == 3 && off == 0)
+				continue;
+                        if (1 <= colons)
+                            *s++ = ':';
+			i = snprintf(s, endp - s, "%02d", (int)(off / 60));
 			if (i < 0) goto err;
 			s += i;
+                        off = off % 60;
+			if (colons == 3 && off == 0)
+				continue;
+                        if (2 <= colons) {
+                            *s++ = ':';
+                            i = snprintf(s, endp - s, "%02d", (int)off);
+                            if (i < 0) goto err;
+                            s += i;
+                        }
 			continue;
 #endif /* MAILHEADER_EXT */
 
@@ -602,34 +618,24 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
 				tp = "UTC";
 				break;
 			}
-#if 0
-#ifdef HAVE_TZNAME
-			i = (daylight && timeptr->tm_isdst > 0); /* 0 or 1 */
-			tp = tzname[i];
-#else
-#ifdef HAVE_TM_ZONE
-			tp = timeptr->tm_zone;
-#else
-#ifdef HAVE_TM_NAME
-			tp = timeptr->tm_name;
-#else
-#ifdef HAVE_TIMEZONE
-			gettimeofday(& tv, & zone);
-#ifdef TIMEZONE_VOID
-			tp = timezone();
-#else
-			tp = timezone(zone.tz_minuteswest, timeptr->tm_isdst > 0);
-#endif /* TIMEZONE_VOID */
-#endif /* HAVE_TIMEZONE */
-#endif /* HAVE_TM_NAME */
-#endif /* HAVE_TM_ZONE */
-#endif /* HAVE_TZNAME */
-#endif /* 0 */
-                        if (vtm->zone == NULL)
-                            tp = "";
-                        else
+			if (vtm->zone == NULL) {
+			    i = 0;
+			}
+			else {
                             tp = vtm->zone;
-			i = strlen(tp);
+			    if (enc) {
+				for (i = 0; i < TBUFSIZE && tp[i]; i++) {
+				    if ((unsigned char)tp[i] > 0x7F) {
+					VALUE str = rb_str_conv_enc_opts(rb_str_new_cstr(tp), rb_locale_encoding(), enc, ECONV_UNDEF_REPLACE|ECONV_INVALID_REPLACE, Qnil);
+					i = strlcpy(tbuf, RSTRING_PTR(str), TBUFSIZE);
+					tp = tbuf;
+					break;
+				    }
+				}
+			    }
+			    else
+				i = strlen(tp);
+			}
 			break;
 
 #ifdef SYSV_EXT
@@ -695,11 +701,13 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
 
 		case 'E':
 			/* POSIX locale extensions, ignored for now */
-			flags |= BIT_OF(LOCALE_E);
+			if (!format[1] || !strchr("cCxXyY", format[1]))
+				goto unknown;
 			goto again;
 		case 'O':
 			/* POSIX locale extensions, ignored for now */
-			flags |= BIT_OF(LOCALE_O);
+			if (!format[1] || !strchr("deHkIlmMSuUVwWy", format[1]))
+				goto unknown;
 			goto again;
 
 		case 'V':	/* week of year according ISO 8601 */
@@ -733,7 +741,13 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
                                         yv = sub(yv, INT2FIX(1));
 
                                 if (*format == 'G') {
-                                        FMTV('0', 1, "d", yv);
+                                        if (FIXNUM_P(yv)) {
+                                                const long y = FIX2LONG(yv);
+                                                FMT('0', 0 <= y ? 4 : 5, "ld", y);
+                                        }
+                                        else {
+                                                FMTV('0', 4, "d", yv);
+                                        }
                                 }
                                 else {
                                         yv = mod(yv, INT2FIX(100));
@@ -799,17 +813,15 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
                                 subsec = div(subsec, INT2FIX(1));
 
                                 if (FIXNUM_P(subsec)) {
-                                        int l;
-                                        l = snprintf(s, endp - s, "%0*ld", precision, FIX2LONG(subsec));
+                                        (void)snprintf(s, endp - s, "%0*ld", precision, FIX2LONG(subsec));
                                         s += precision;
                                 }
                                 else {
                                         VALUE args[2], result;
-                                        size_t l;
                                         args[0] = INT2FIX(precision);
                                         args[1] = subsec;
                                         result = rb_str_format(2, args, rb_str_new2("%0*d"));
-                                        l = strlcpy(s, StringValueCStr(result), endp-s);
+                                        (void)strlcpy(s, StringValueCStr(result), endp-s);
                                         s += precision;
                                 }
 			}
@@ -840,14 +852,26 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
 			padding = ' ';
 			goto again;
 
+		case ':':
+			for (colons = 1; colons <= 3; ++colons) {
+				if (format+colons >= format_end) goto unknown;
+				if (format[colons] == 'z') break;
+				if (format[colons] != ':') goto unknown;
+			}
+			format += colons - 1;
+			goto again;
+
 		case '0':
 			padding = '0';
 		case '1':  case '2': case '3': case '4':
 		case '5': case '6':  case '7': case '8': case '9':
 			{
-				char *e;
-				precision = (int)strtoul(format, &e, 10);
-				format = e - 1;
+				size_t n;
+				int ov;
+				unsigned long u = ruby_scan_digits(format, format_end-format, 10, &n, &ov);
+				if (ov || u > INT_MAX) goto unknown;
+				precision = (int)u;
+				format += n - 1;
 				goto again;
 			}
 
@@ -858,60 +882,69 @@ rb_strftime_with_timespec(char *s, size_t maxsize, const char *format, const str
 			precision = -1;
 			flags = 0;
 			padding = 0;
+                        colons = 0;
 			break;
 		}
 		if (i) {
 			FILL_PADDING(i);
 			memcpy(s, tp, i);
-			switch (flags & (BIT_OF(UPPER)|BIT_OF(LOWER))) {
-			case BIT_OF(UPPER):
-				do {
-					if (ISLOWER(*s)) *s = TOUPPER(*s);
-				} while (s++, --i);
-				break;
-			case BIT_OF(LOWER):
-				do {
-					if (ISUPPER(*s)) *s = TOLOWER(*s);
-				} while (s++, --i);
-				break;
-			default:
-				s += i;
-				break;
-			}
+			s = case_conv(s, i, flags);
 		}
 	}
-	if (s >= endp) {
-		goto err;
-	}
-	if (*format == '\0') {
-		*s = '\0';
-		return (s - start);
-	} else
+	if (format != format_end) {
 		return 0;
+	}
+	len = s - start;
+	rb_str_set_len(ftime, len);
+	rb_str_resize(ftime, len);
+	return ftime;
 }
 
-size_t
-rb_strftime(char *s, size_t maxsize, const char *format, const struct vtm *vtm, VALUE timev, int gmt)
+static size_t
+strftime_size_limit(size_t format_len)
 {
-    return rb_strftime_with_timespec(s, maxsize, format, vtm, timev, NULL, gmt);
+	size_t limit = format_len * (1*1024*1024);
+	if (limit < format_len) limit = format_len;
+	else if (limit < 1024) limit = 1024;
+	return limit;
 }
 
-size_t
-rb_strftime_timespec(char *s, size_t maxsize, const char *format, const struct vtm *vtm, struct timespec *ts, int gmt)
+VALUE
+rb_strftime(const char *format, size_t format_len,
+	    rb_encoding *enc, const struct vtm *vtm, VALUE timev, int gmt)
 {
-    return rb_strftime_with_timespec(s, maxsize, format, vtm, Qnil, ts, gmt);
+	VALUE result = rb_enc_str_new(0, 0, enc);
+	return rb_strftime_with_timespec(result, format, format_len, enc,
+					 vtm, timev, NULL, gmt,
+					 strftime_size_limit(format_len));
 }
 
+VALUE
+rb_strftime_timespec(const char *format, size_t format_len,
+		     rb_encoding *enc, const struct vtm *vtm, struct timespec *ts, int gmt)
+{
+	VALUE result = rb_enc_str_new(0, 0, enc);
+	return rb_strftime_with_timespec(result, format, format_len, enc,
+					 vtm, Qnil, ts, gmt,
+					 strftime_size_limit(format_len));
+}
+
+#if 0
+VALUE
+rb_strftime_limit(const char *format, size_t format_len,
+		  rb_encoding *enc, const struct vtm *vtm, struct timespec *ts,
+		  int gmt, size_t maxsize)
+{
+	VALUE result = rb_enc_str_new(0, 0, enc);
+	return rb_strftime_with_timespec(result, format, format_len, enc,
+					 vtm, Qnil, ts, gmt, maxsize);
+}
+#endif
+
 /* isleap --- is a year a leap year? */
 
-#ifndef __STDC__
-static int
-isleap(year)
-long year;
-#else
 static int
 isleap(long year)
-#endif
 {
 	return ((year % 4 == 0 && year % 100 != 0) || year % 400 == 0);
 }
@@ -945,14 +978,8 @@ vtm2tm_noyear(const struct vtm *vtm, struct tm *result)
 #ifdef POSIX2_DATE
 /* iso8601wknum --- compute week number according to ISO 8601 */
 
-#ifndef __STDC__
-static int
-iso8601wknum(timeptr)
-const struct tm *timeptr;
-#else
 static int
 iso8601wknum(const struct tm *timeptr)
-#endif
 {
 	/*
 	 * From 1003.2:
@@ -1072,15 +1099,8 @@ iso8601wknum_v(const struct vtm *vtm)
 
 /* With thanks and tip of the hatlo to ado@elsie.nci.nih.gov */
 
-#ifndef __STDC__
-static int
-weeknumber(timeptr, firstweekday)
-const struct tm *timeptr;
-int firstweekday;
-#else
 static int
 weeknumber(const struct tm *timeptr, int firstweekday)
-#endif
 {
 	int wday = timeptr->tm_wday;
 	int ret;
@@ -1216,9 +1236,7 @@ static char *array[] =
 /* main routine. */
 
 int
-main(argc, argv)
-int argc;
-char **argv;
+main(int argc, char **argv)
 {
 	long time();
 
