@@ -1,4 +1,5 @@
 # coding: US-ASCII
+# frozen_string_literal: false
 
 require 'rdoc/test_case'
 
@@ -136,9 +137,67 @@ class TestRDocGeneratorJsonIndex < RDoc::TestCase
     assert_equal expected, index
   end
 
-  def test_generate_utf_8
-    skip "Encoding not implemented" unless Object.const_defined? :Encoding
+  def test_generate_gzipped
+    begin
+      require 'zlib'
+    rescue LoadError
+      skip "no zlib"
+    end
+    @g.generate
+    @g.generate_gzipped
 
+    assert_file 'js/searcher.js'
+    assert_file 'js/searcher.js.gz'
+    assert_file 'js/navigation.js'
+    assert_file 'js/navigation.js.gz'
+    assert_file 'js/search_index.js'
+    assert_file 'js/search_index.js.gz'
+
+    json = File.open('js/search_index.js.gz', 'rb') {|gzip|
+      Zlib::GzipReader.new(gzip).read
+    }
+
+    json =~ /\Avar search_data = /
+
+    assignment = $&
+    index = $'
+
+    refute_empty assignment
+
+    index = JSON.parse index
+
+    info = [
+      @klass.search_record[2..-1],
+      @nest_klass.search_record[2..-1],
+      @meth.search_record[2..-1],
+      @nest_meth.search_record[2..-1],
+      @page.search_record[2..-1],
+    ]
+
+    expected = {
+      'index' => {
+        'searchIndex' => [
+          'c',
+          'd',
+          'meth()',
+          'meth()',
+          'page',
+        ],
+        'longSearchIndex' => [
+          'c',
+          'c::d',
+          'c#meth()',
+          'c::d#meth()',
+          '',
+        ],
+        'info' => info,
+      },
+    }
+
+    assert_equal expected, index
+  end
+
+  def test_generate_utf_8
     text = "5\xB0"
     text.force_encoding Encoding::ISO_8859_1
     @klass.add_comment comment(text), @top_level
@@ -261,4 +320,3 @@ class TestRDocGeneratorJsonIndex < RDoc::TestCase
   end
 
 end
-
