@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 #
 # test_scanner_events.rb
 #
@@ -38,6 +39,10 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  Ripper.tokenize("print( <<""EOS)\nheredoc\nEOS\n")
     assert_equal ["\#\n", "\n", "\#\n", "\n", "nil", "\n"],
                  Ripper.tokenize("\#\n\n\#\n\nnil\n")
+    assert_equal ["1", "  ", ".", "foo", "\n"],
+                 Ripper.tokenize("1  .foo\n")
+    assert_equal ["1", "\n", "  ", ".", "foo", "\n"],
+                 Ripper.tokenize("1\n  .foo\n")
   end
 
   def test_lex
@@ -98,6 +103,12 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                   [[5, 0], :on_imaginary, "5.6ri"],
                  ],
                  Ripper.lex("1r\n2i\n3ri\n4.2r\n5.6ri")
+     assert_equal [[[1, 0], :on_heredoc_beg, "<<~EOS"],
+                   [[1, 6], :on_nl, "\n"],
+                   [[2, 2], :on_tstring_content, "heredoc\n"],
+                   [[3, 0], :on_heredoc_end, "EOS"]
+                 ],
+                 Ripper.lex("<<~EOS\n  heredoc\nEOS")
   end
 
   def test_location
@@ -531,6 +542,8 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
                  scan('op', ':[]')
     assert_equal ['[]='],
                  scan('op', ':[]=')
+    assert_equal ['&.'],
+                 scan('op', 'a&.f')
     assert_equal [],
                  scan('op', %q[`make all`])
   end
@@ -878,15 +891,28 @@ class TestRipper::ScannerEvents < Test::Unit::TestCase
   end
 
   def test_label
+    assert_equal %w(foo:),
+                 scan('label', '{foo: 1}')
+  end
+
+  def test_label_end
+    assert_equal %w(":),
+                 scan('label_end', '{"foo-bar": 1}')
   end
 
   def test_tlambda
+    assert_equal %w(->),
+                 scan('tlambda', '->{}')
   end
 
   def test_tlambeg
+    assert_equal %w({),
+                 scan('tlambeg', '-> {}')
   end
 
   def test_tlambda_arg
+    assert_equal %w(),
+                 scan('tlambda_arg', '-> {}')
   end
 
 end if ripper_test
