@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 require 'test/unit'
 require 'tmpdir'
 
@@ -20,6 +21,7 @@ class TestDir_M17N < Test::Unit::TestCase
         assert_include(ents, filename)
       EOS
 
+      return if /cygwin/ =~ RUBY_PLATFORM
       assert_separately(%w[-EASCII-8BIT], <<-EOS, :chdir=>dir)
         filename = #{code}.chr('UTF-8').force_encoding("ASCII-8BIT")
         opts = {:encoding => Encoding.default_external} if /mswin|mingw/ =~ RUBY_PLATFORM
@@ -56,6 +58,7 @@ class TestDir_M17N < Test::Unit::TestCase
   end
 
   def test_filename_extutf8_invalid
+    return if /cygwin/ =~ RUBY_PLATFORM
     with_tmpdir {|d|
       assert_separately(%w[-EASCII-8BIT], <<-'EOS', :chdir=>d)
         filename = "\xff".force_encoding("ASCII-8BIT") # invalid byte sequence as UTF-8
@@ -171,6 +174,7 @@ class TestDir_M17N < Test::Unit::TestCase
   ## others
 
   def test_filename_bytes_euc_jp
+    return if /cygwin/ =~ RUBY_PLATFORM
     with_tmpdir {|d|
       assert_separately(%w[-EEUC-JP], <<-'EOS', :chdir=>d)
         filename = "\xA4\xA2".force_encoding("euc-jp")
@@ -187,6 +191,7 @@ class TestDir_M17N < Test::Unit::TestCase
   end
 
   def test_filename_euc_jp
+    return if /cygwin/ =~ RUBY_PLATFORM
     with_tmpdir {|d|
       assert_separately(%w[-EEUC-JP], <<-'EOS', :chdir=>d)
         filename = "\xA4\xA2".force_encoding("euc-jp")
@@ -208,7 +213,7 @@ class TestDir_M17N < Test::Unit::TestCase
           when /darwin/
             filename = filename.encode("utf-8", "euc-jp").b
           when /mswin|mingw/
-            if ents.include?(win_expected_filename.dup.force_encoding("ASCII-8BIT"))
+            if ents.include?(win_expected_filename.b)
               ents = Dir.entries(".", {:encoding => Encoding.find("filesystem")})
               filename = win_expected_filename
             end
@@ -232,6 +237,7 @@ class TestDir_M17N < Test::Unit::TestCase
   end
 
   def test_filename_ext_euc_jp_and_int_utf_8
+    return if /cygwin/ =~ RUBY_PLATFORM
     with_tmpdir {|d|
       assert_separately(%w[-EEUC-JP], <<-'EOS', :chdir=>d)
         filename = "\xA4\xA2".force_encoding("euc-jp")
@@ -303,20 +309,104 @@ class TestDir_M17N < Test::Unit::TestCase
     }
   end
 
+  def with_enc_path
+    with_tmpdir do |d|
+      names = %W"\u{391 392 393 394 395} \u{3042 3044 3046 3048 304a}"
+      names.each do |dir|
+        EnvUtil.with_default_external(Encoding::UTF_8) do
+          Dir.mkdir(dir) rescue next
+          begin
+            yield(dir)
+          ensure
+            File.chmod(0700, dir)
+          end
+        end
+      end
+    end
+  end
+
+  def test_glob_warning_opendir
+    with_enc_path do |dir|
+      open("#{dir}/x", "w") {}
+      File.chmod(0300, dir)
+      next if File.readable?(dir)
+      assert_warning(/#{dir}/) do
+        Dir.glob("#{dir}/*")
+      end
+    end
+  end
+
+  def test_glob_warning_match_all
+    with_enc_path do |dir|
+      open("#{dir}/x", "w") {}
+      File.chmod(0000, dir)
+      next if File.readable?(dir)
+      assert_warning(/#{dir}/) do
+        Dir.glob("#{dir}/x")
+      end
+    end
+  end
+
+  def test_glob_warning_match_dir
+    with_enc_path do |dir|
+      Dir.mkdir("#{dir}/x")
+      File.chmod(0000, dir)
+      next if File.readable?(dir)
+      assert_warning(/#{dir}/) do
+        Dir.glob("#{dir}/x/")
+      end
+    end
+  end
+
+  def test_glob_escape_multibyte
+    name = "\x81\\".force_encoding(Encoding::Shift_JIS)
+    with_tmpdir do
+      open(name, "w") {} rescue next
+      match, = Dir.glob("#{name}*")
+      next unless match and match.encoding == Encoding::Shift_JIS
+      assert_equal([name], Dir.glob("\\#{name}*"))
+    end
+  end
+
+  def test_glob_encoding
+    with_tmpdir do
+      list = %W"file_one.ext file_two.ext \u{6587 4ef6}1.txt \u{6587 4ef6}2.txt"
+      list.each {|f| open(f, "w") {}}
+      a = "file_one*".force_encoding Encoding::IBM437
+      b = "file_two*".force_encoding Encoding::EUC_JP
+      assert_equal([a, b].map(&:encoding), Dir[a, b].map(&:encoding))
+      dir = "\u{76EE 5F551}"
+      Dir.mkdir(dir)
+      list << dir
+      bug12081 = '[ruby-core:73868] [Bug #12081]'
+      a = "*".force_encoding("us-ascii")
+      result = Dir[a].map {|n|
+        if n.encoding == Encoding::ASCII_8BIT ||
+            n.encoding == Encoding::ISO_8859_1 ||
+            !n.valid_encoding?
+          n.force_encoding(Encoding::UTF_8)
+        else
+          n.encode(Encoding::UTF_8)
+        end
+      }
+      assert_equal(list, result.sort!, bug12081)
+    end
+  end
+
+  PP = Object.new.extend(Test::Unit::Assertions)
+  def PP.mu_pp(ary) #:nodoc:
+    '[' << ary.map {|str| "#{str.dump}(#{str.encoding})"}.join(', ') << ']'
+  end
+
   def test_entries_compose
     bug7267 = '[ruby-core:48745] [Bug #7267]'
 
-    pp = Object.new.extend(Test::Unit::Assertions)
-    def pp.mu_pp(ary) #:nodoc:
-      '[' << ary.map {|str| "#{str.dump}(#{str.encoding})"}.join(', ') << ']'
-    end
-
     with_tmpdir {|d|
       orig = %W"d\u{e9}tente x\u{304c 304e 3050 3052 3054}"
       orig.each {|n| open(n, "w") {}}
       if /mswin|mingw/ =~ RUBY_PLATFORM
         opts = {:encoding => Encoding.default_external}
-        orig.map! {|o| o.encode(Encoding.find("filesystem")) rescue o.tr("^a-z", "?")}
+        orig.map! {|o| o.encode("filesystem") rescue o.tr("^a-z", "?")}
       else
         enc = Encoding.find("filesystem")
         enc = Encoding::ASCII_8BIT if enc == Encoding::US_ASCII
@@ -324,7 +414,28 @@ class TestDir_M17N < Test::Unit::TestCase
       end
       ents = Dir.entries(".", opts).reject {|n| /\A\./ =~ n}
       ents.sort!
-      pp.assert_equal(orig, ents, bug7267)
+      PP.assert_equal(orig, ents, bug7267)
     }
   end
+
+  def test_pwd
+    orig = %W"d\u{e9}tente x\u{304c 304e 3050 3052 3054}"
+    expected = []
+    results = []
+    orig.each {|o|
+      if /mswin|mingw/ =~ RUBY_PLATFORM
+        n = (o.encode("filesystem") rescue next)
+      else
+        enc = Encoding.find("filesystem")
+        enc = Encoding::ASCII_8BIT if enc == Encoding::US_ASCII
+        n = o.dup.force_encoding(enc)
+      end
+      expected << n
+      with_tmpdir {
+        Dir.mkdir(o)
+        results << File.basename(Dir.chdir(o) {Dir.pwd})
+      }
+    }
+    PP.assert_equal(expected, results)
+  end
 end
