@@ -1,5 +1,5 @@
+# frozen_string_literal: false
 require 'test/unit'
-require_relative 'envutil'
 
 class TestEnumerator < Test::Unit::TestCase
   def setup
@@ -25,7 +25,7 @@ class TestEnumerator < Test::Unit::TestCase
   def test_iterators
     assert_equal [0, 1, 2], enum_test(3.times)
     assert_equal [:x, :y, :z], enum_test([:x, :y, :z].each)
-    assert_equal [[:x, 1], [:y, 2]], enum_test({:x=>1, :y=>2})
+    assert_equal [[:x, 1], [:y, 2]], enum_test({:x=>1, :y=>2}.each)
   end
 
   ## Enumerator as Iterator
@@ -47,6 +47,14 @@ class TestEnumerator < Test::Unit::TestCase
     }
   end
 
+  def test_loop_return_value
+    assert_equal nil, loop { break }
+    assert_equal 42,  loop { break 42 }
+
+    e = Enumerator.new { |y| y << 1; y << 2; :stopped }
+    assert_equal :stopped, loop { e.next while true }
+  end
+
   def test_nested_iteration
     def (o = Object.new).each
       yield :ok1
@@ -95,6 +103,7 @@ class TestEnumerator < Test::Unit::TestCase
       1.times do
         foo = [1,2,3].to_enum
         GC.start
+        foo
       end
       GC.start
     end
@@ -434,6 +443,18 @@ class TestEnumerator < Test::Unit::TestCase
     assert_raise(RuntimeError) {
       g.__send__ :initialize, proc { |y| y << 4 << 5 }
     }
+
+    g = Enumerator::Generator.new(proc {|y| y << 4 << 5; :foo })
+    a = []
+    assert_equal(:foo, g.each {|x| a << x })
+    assert_equal([4, 5], a)
+
+    assert_raise(LocalJumpError) {Enumerator::Generator.new}
+    assert_raise(TypeError) {Enumerator::Generator.new(1)}
+    obj = eval("class C\u{1f5ff}; self; end").new
+    assert_raise_with_message(TypeError, /C\u{1f5ff}/) {
+      Enumerator::Generator.new(obj)
+    }
   end
 
   def test_generator_args
@@ -625,5 +646,10 @@ class TestEnumerator < Test::Unit::TestCase
     e.next
     assert_raise(StopIteration) { e.peek }
   end
+
+  def test_uniq
+    assert_equal([1, 2, 3, 4, 5, 10],
+                 (1..Float::INFINITY).lazy.uniq{|x| (x**2) % 10 }.first(6))
+  end
 end
 
