@@ -35,9 +35,9 @@ class TestSetTraceFunc < Test::Unit::TestCase
                  events.shift)
     assert_equal(["line", 4, __method__, self.class],
                  events.shift)
-    assert_equal(["c-call", 4, :+, Fixnum],
+    assert_equal(["c-call", 4, :+, Integer],
                  events.shift)
-    assert_equal(["c-return", 4, :+, Fixnum],
+    assert_equal(["c-return", 4, :+, Integer],
                  events.shift)
     assert_equal(["line", 5, __method__, self.class],
                  events.shift)
@@ -73,9 +73,9 @@ class TestSetTraceFunc < Test::Unit::TestCase
                  events.shift)
     assert_equal(["line", 5, :add, self.class],
                  events.shift)
-    assert_equal(["c-call", 5, :+, Fixnum],
+    assert_equal(["c-call", 5, :+, Integer],
                  events.shift)
-    assert_equal(["c-return", 5, :+, Fixnum],
+    assert_equal(["c-return", 5, :+, Integer],
                  events.shift)
     assert_equal(["return", 6, :add, self.class],
                  events.shift)
@@ -353,8 +353,8 @@ class TestSetTraceFunc < Test::Unit::TestCase
      ["c-return", 8, :new, Class],
      ["call", 4, :foo, ThreadTraceInnerClass],
      ["line", 5, :foo, ThreadTraceInnerClass],
-     ["c-call", 5, :+, Fixnum],
-     ["c-return", 5, :+, Fixnum],
+     ["c-call", 5, :+, Integer],
+     ["c-return", 5, :+, Integer],
      ["return", 6, :foo, ThreadTraceInnerClass],
      ["line", 9, __method__, self.class],
      ["c-call", 9, :set_trace_func, Thread]].each do |e|
@@ -768,10 +768,10 @@ class TestSetTraceFunc < Test::Unit::TestCase
     # pp events
     # expected_events =
     [[:b_call, :test_tracepoint_block, TestSetTraceFunc, TestSetTraceFunc, nil],
-     [:c_call, :times, Integer, Fixnum, nil],
+     [:c_call, :times, Integer, Integer, nil],
      [:b_call, :test_tracepoint_block, TestSetTraceFunc, TestSetTraceFunc, nil],
      [:b_return, :test_tracepoint_block, TestSetTraceFunc, TestSetTraceFunc, 3],
-     [:c_return, :times, Integer, Fixnum, 1],
+     [:c_return, :times, Integer, Integer, 1],
      [:call, :method_for_test_tracepoint_block, TestSetTraceFunc, TestSetTraceFunc, nil],
      [:b_call, :test_tracepoint_block, TestSetTraceFunc, TestSetTraceFunc, nil],
      [:b_return, :test_tracepoint_block, TestSetTraceFunc, TestSetTraceFunc, 4],
@@ -1359,7 +1359,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     assert_consistent_call_return '[Bug #9959]' do
       begin
         method_test_argument_error_on_bmethod(wrong_key: 2)
-      rescue => e
+      rescue
         # ignore
       end
     end
@@ -1369,7 +1369,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     assert_consistent_call_return '[Bug #9961]' do
       begin
         -Numeric.new
-      rescue => e
+      rescue
         # ignore
       end
     end
@@ -1432,7 +1432,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
       return if Thread.current != target_th
       evs << tp.event
     }.enable{
-      a = Bug10724.new
+      Bug10724.new
     }
 
     assert_equal([:call, :return], evs)
@@ -1489,149 +1489,114 @@ class TestSetTraceFunc < Test::Unit::TestCase
     }
   end
 
-  # tests for `return_value` with non-local exit [Bug #13369]
-
-  def tp_return_value mid
-    ary = []
-    TracePoint.new(:return, :b_return){|tp| ary << [tp.event, tp.method_id, tp.return_value]}.enable{
-      send mid
+  def test_tracepoint_callee_id
+    events = []
+    capture_events = Proc.new{|tp|
+      events << [tp.event, tp.method_id, tp.callee_id]
     }
-    ary.pop # last b_return event is not required.
-    ary
-  end
 
-  def f_raise
-    raise
-  rescue
-    return :f_raise_return
-  end
-
-  def f_iter1
-    yield
-    return :f_iter1_return
-  end
-
-  def f_iter2
-    yield
-    return :f_iter2_return
-  end
-
-  def f_return_in_iter
-    f_iter1 do
-      f_iter2 do
-        return :f_return_in_iter_return
-      end
-    end
-    2
-  end
-
-  def f_break_in_iter
-    f_iter1 do
-      f_iter2 do
-        break :f_break_in_iter_break
-      end
-      :f_iter1_block_value
-    end
-    :f_break_in_iter_return
-  end
-
-  def test_return_value_with_rescue
-    assert_equal [[:return,   :f_raise,          :f_raise_return]],
-                 tp_return_value(:f_raise),
-                 '[Bug #13369]'
-
-    assert_equal [[:b_return, :f_return_in_iter, nil],
-                  [:return,   :f_iter2,          nil],
-                  [:b_return, :f_return_in_iter, nil],
-                  [:return,   :f_iter1,          nil],
-                  [:return,   :f_return_in_iter, :f_return_in_iter_return]],
-                 tp_return_value(:f_return_in_iter),
-                 '[Bug #13369]'
-
-    assert_equal [[:b_return, :f_break_in_iter,  :f_break_in_iter_break],
-                  [:return,   :f_iter2,          nil],
-                  [:b_return, :f_break_in_iter,  :f_iter1_block_value],
-                  [:return,   :f_iter1,          :f_iter1_return],
-                  [:return,   :f_break_in_iter,  :f_break_in_iter_return]],
-                 tp_return_value(:f_break_in_iter),
-                 '[Bug #13369]'
-  end
-
-  define_method(:f_last_defined) do
-    :f_last_defined
-  end
-
-  define_method(:f_return_defined) do
-    return :f_return_defined
-  end
-
-  define_method(:f_break_defined) do
-    return :f_break_defined
-  end
-
-  define_method(:f_raise_defined) do
-    begin
-      raise
-    rescue
-      return :f_raise_defined
-    end
-  end
-
-  define_method(:f_break_in_rescue_defined) do
-    begin
-      raise
-    rescue
-      break :f_break_in_rescue_defined
-    end
-  end
-
-  def test_return_value_with_rescue_and_defined_methods
-    assert_equal [[:b_return, :f_last_defined, :f_last_defined],
-                  [:return,   :f_last_defined, :f_last_defined]],
-                 tp_return_value(:f_last_defined),
-                 '[Bug #13369]'
-
-    assert_equal [[:b_return, :f_return_defined, nil], # current limitation
-                  [:return,   :f_return_defined, :f_return_defined]],
-                 tp_return_value(:f_return_defined),
-                 '[Bug #13369]'
-
-    assert_equal [[:b_return, :f_break_defined, nil],
-                  [:return,   :f_break_defined, :f_break_defined]],
-                 tp_return_value(:f_break_defined),
-                 '[Bug #13369]'
-
-    assert_equal [[:b_return, :f_raise_defined, nil],
-                  [:return,   :f_raise_defined, f_raise_defined]],
-                 tp_return_value(:f_raise_defined),
-                 '[Bug #13369]'
-
-    assert_equal [[:b_return, :f_break_in_rescue_defined, nil],
-                  [:return,   :f_break_in_rescue_defined, f_break_in_rescue_defined]],
-                 tp_return_value(:f_break_in_rescue_defined),
-                 '[Bug #13369]'
-  end
-
-  def f_iter
-    yield
-  end
-
-  def f_break_in_rescue
-    f_iter do
-      begin
+    o = Class.new{
+      def m
         raise
-      rescue
-        break :b
       end
-    end
-    :f_break_in_rescue_return_value
-  end
+      alias alias_m m
+    }.new
+    TracePoint.new(:raise, :call, :return, &capture_events).enable{
+      o.alias_m rescue nil
+    }
+    assert_equal [[:call, :m, :alias_m], [:raise, :m, :alias_m], [:return, :m, :alias_m]], events
+    events.clear
 
-  def test_break_with_rescue
-    assert_equal [[:b_return, :f_break_in_rescue, :b],
-                  [:return, :f_iter, nil],
-                  [:return, :f_break_in_rescue, :f_break_in_rescue_return_value]],
-                 tp_return_value(:f_break_in_rescue),
-                 '[Bug #13369]'
+    o = Class.new{
+      alias alias_raise raise
+      def m
+        alias_raise
+      end
+    }.new
+    TracePoint.new(:c_return, &capture_events).enable{
+      o.m rescue nil
+    }
+    assert_equal [:c_return, :raise, :alias_raise], events[0]
+    events.clear
+
+    o = Class.new(String){
+      include Enumerable
+      alias each each_char
+    }.new('foo')
+    TracePoint.new(:c_return, &capture_events).enable{
+      o.find{true}
+    }
+    assert_equal [:c_return, :each_char, :each], events[0]
+    events.clear
+
+    o = Class.new{
+      define_method(:m){}
+      alias alias_m m
+    }.new
+    TracePoint.new(:call, :return, &capture_events).enable{
+      o.alias_m
+    }
+    assert_equal [[:call, :m, :alias_m], [:return, :m, :alias_m]], events
+    events.clear
+
+    o = Class.new{
+      def m
+        tap{return}
+      end
+      alias alias_m m
+    }.new
+    TracePoint.new(:return, &capture_events).enable{
+      o.alias_m
+    }
+    assert_equal [[:return, :m, :alias_m]], events
+    events.clear
+
+    o = Class.new{
+      define_method(:m){raise}
+      alias alias_m m
+    }.new
+    TracePoint.new(:b_return, :return, &capture_events).enable{
+      o.alias_m rescue nil
+    }
+    assert_equal [[:b_return, :m, :alias_m], [:return, :m, :alias_m]], events[0..1]
+    events.clear
+
+    o = Class.new{
+      define_method(:m){tap{return}}
+      alias alias_m m
+    }.new
+    TracePoint.new(:b_return, &capture_events).enable{
+      o.alias_m
+    }
+    assert_equal [[:b_return, :m, :alias_m], [:b_return, :m, :alias_m]], events[0..1]
+    events.clear
+
+    o = Class.new{
+      alias alias_tap tap
+      define_method(:m){alias_tap{return}}
+    }.new
+    TracePoint.new(:c_return, &capture_events).enable{
+      o.m
+    }
+    assert_equal [[:c_return, :tap, :alias_tap]], events
+    events.clear
+
+    c = Class.new{
+      alias initialize itself
+    }
+    TracePoint.new(:c_call, &capture_events).enable{
+      c.new
+    }
+    assert_equal [:c_call, :itself, :initialize], events[1]
+    events.clear
+
+    o = Class.new{
+      alias alias_itself itself
+    }.new
+    TracePoint.new(:c_call, :c_return, &capture_events).enable{
+      o.alias_itself
+    }
+    assert_equal [[:c_call, :itself, :alias_itself], [:c_return, :itself, :alias_itself]], events
+    events.clear
   end
 end
