@@ -93,6 +93,20 @@ class TestSyntax < Test::Unit::TestCase
     assert_valid_syntax("tap (proc do end)", __FILE__, bug9726)
   end
 
+  def test_normal_argument
+    assert_valid_syntax('def foo(x) end')
+    assert_syntax_error('def foo(X) end', /constant/)
+    assert_syntax_error('def foo(@x) end', /instance variable/)
+    assert_syntax_error('def foo(@@x) end', /class variable/)
+  end
+
+  def test_optional_argument
+    assert_valid_syntax('def foo(x=nil) end')
+    assert_syntax_error('def foo(X=nil) end', /constant/)
+    assert_syntax_error('def foo(@x=nil) end', /instance variable/)
+    assert_syntax_error('def foo(@@x=nil) end', /class variable/)
+  end
+
   def test_keyword_rest
     bug5989 = '[ruby-core:42455]'
     assert_valid_syntax("def kwrest_test(**a) a; end", __FILE__, bug5989)
@@ -155,16 +169,10 @@ class TestSyntax < Test::Unit::TestCase
   end
 
   def test_keyword_empty_splat
-    assert_separately([], "#{<<-"begin;"}\n#{<<-'end;'}")
-    begin;
+    assert_separately([], <<-'end;')
       bug10719 = '[ruby-core:67446] [Bug #10719]'
       assert_valid_syntax("foo(a: 1, **{})", bug10719)
     end;
-    assert_separately([], "#{<<-"begin;"}\n#{<<-'end;'}")
-    begin;
-      bug13756 = '[ruby-core:82113] [Bug #13756]'
-      assert_valid_syntax("defined? foo(**{})", bug13756)
-    end;
   end
 
   def test_keyword_self_reference
@@ -211,11 +219,16 @@ class TestSyntax < Test::Unit::TestCase
   def test_keyword_invalid_name
     bug11663 = '[ruby-core:71356] [Bug #11663]'
 
-    o = Object.new
+    o = o = Object.new
     assert_syntax_error('def o.foo(arg1?:) end', /arg1\?/, bug11663)
     assert_syntax_error('def o.foo(arg1?:, arg2:) end', /arg1\?/, bug11663)
     assert_syntax_error('proc {|arg1?:|}', /arg1\?/, bug11663)
     assert_syntax_error('proc {|arg1?:, arg2:|}', /arg1\?/, bug11663)
+
+    bug10545 = '[ruby-dev:48742] [Bug #10545]'
+    assert_syntax_error('def o.foo(FOO: a) end', /constant/, bug10545)
+    assert_syntax_error('def o.foo(@foo: a) end', /instance variable/)
+    assert_syntax_error('def o.foo(@@foo: a) end', /class variable/)
   end
 
   def test_optional_self_reference
@@ -298,8 +311,14 @@ WARN
      [:/, "regexp literal"],
      [:%, "string literal"],
     ].each do |op, syn|
-      assert_warning(warning % [op, syn]) do
-        assert_valid_syntax("puts 1 #{op}0", "test", verbose: true)
+      all_assertions do |a|
+        ["puts 1 #{op}0", "puts :a #{op}0", "m = 1; puts m #{op}0"].each do |src|
+          a.for(src) do
+            assert_warning(warning % [op, syn], src) do
+              assert_valid_syntax(src, "test", verbose: true)
+            end
+          end
+        end
       end
     end
   end
@@ -368,29 +387,29 @@ WARN
 
   def test_cmdarg_kwarg_lvar_clashing_method
     bug12073 = '[ruby-core:73816] [Bug#12073]'
-    a = 1
+    a = a = 1
     assert_valid_syntax("a b: 1")
     assert_valid_syntax("a = 1; a b: 1", bug12073)
   end
 
   def test_duplicated_arg
     assert_syntax_error("def foo(a, a) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(_, _) end }
+    assert_valid_syntax("def foo(_, _) end")
   end
 
   def test_duplicated_rest
     assert_syntax_error("def foo(a, *a) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(_, *_) end }
+    assert_valid_syntax("def foo(_, *_) end")
   end
 
   def test_duplicated_opt
     assert_syntax_error("def foo(a, a=1) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(_, _=1) end }
+    assert_valid_syntax("def foo(_, _=1) end")
   end
 
   def test_duplicated_opt_rest
     assert_syntax_error("def foo(a=1, *a) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(_=1, *_) end }
+    assert_valid_syntax("def foo(_=1, *_) end")
   end
 
   def test_duplicated_rest_opt
@@ -403,12 +422,12 @@ WARN
 
   def test_duplicated_opt_post
     assert_syntax_error("def foo(a=1, a) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(_=1, _) end }
+    assert_valid_syntax("def foo(_=1, _) end")
   end
 
   def test_duplicated_kw
     assert_syntax_error("def foo(a, a: 1) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(_, _: 1) end }
+    assert_valid_syntax("def foo(_, _: 1) end")
   end
 
   def test_duplicated_rest_kw
@@ -418,22 +437,22 @@ WARN
 
   def test_duplicated_opt_kw
     assert_syntax_error("def foo(a=1, a: 1) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(_=1, _: 1) end }
+    assert_valid_syntax("def foo(_=1, _: 1) end")
   end
 
   def test_duplicated_kw_kwrest
     assert_syntax_error("def foo(a: 1, **a) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(_: 1, **_) end }
+    assert_valid_syntax("def foo(_: 1, **_) end")
   end
 
   def test_duplicated_rest_kwrest
     assert_syntax_error("def foo(*a, **a) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(*_, **_) end }
+    assert_valid_syntax("def foo(*_, **_) end")
   end
 
   def test_duplicated_opt_kwrest
     assert_syntax_error("def foo(a=1, **a) end", /duplicated argument name/)
-    assert_nothing_raised { def foo(_=1, **_) end }
+    assert_valid_syntax("def foo(_=1, **_) end")
   end
 
   def test_duplicated_when
@@ -448,7 +467,7 @@ WARN
       }
     }
     assert_warning(/#{w}/){#/3: #{w}.+4: #{w}.+5: #{w}.+5: #{w}/m){
-      a = 1
+      a = a = 1
       eval %q{
         case 1
         when 1, 1
@@ -459,16 +478,8 @@ WARN
     }
   end
 
-  def test_invalid_break
-    assert_syntax_error("def m; break; end", /Invalid break/)
-    assert_in_out_err([], '/#{break}/', [],  /Invalid break \(SyntaxError\)$/)
-    assert_in_out_err([], '/#{break}/o', [],  /Invalid break \(SyntaxError\)$/)
-  end
-
   def test_invalid_next
     assert_syntax_error("def m; next; end", /Invalid next/)
-    assert_in_out_err([], '/#{next}/', [],  /Invalid next \(SyntaxError\)$/)
-    assert_in_out_err([], '/#{next}/o', [],  /Invalid next \(SyntaxError\)$/)
   end
 
   def test_lambda_with_space
@@ -604,7 +615,7 @@ e"
   end
 
   def test_dedented_heredoc_with_interpolated_string
-    w = ""
+    w = w = ""
     result = " \#{mesg} a\n" \
              "  zy\n"
     expect = '#{mesg} a'"\n" \
@@ -738,6 +749,12 @@ eom
     assert_syntax_error("puts <<""EOS\n""ng\n""EOS\r""NO\n", /can't find string "EOS" anywhere before EOF/)
   end
 
+  def test_heredoc_newline
+    assert_warn(/ends with a newline/) do
+      eval("<<\"EOS\n\"\nEOS\n")
+    end
+  end
+
   def test__END___cr
     assert_syntax_error("__END__\r<<<<<\n", /unexpected <</)
   end
@@ -764,12 +781,6 @@ eom
     assert_syntax_error(code, /def n "\u{2208}"; end/, bug10114)
   end
 
-  def test_bad_kwarg
-    bug10545 = '[ruby-dev:48742] [Bug #10545]'
-    src = 'def foo(A: a) end'
-    assert_syntax_error(src, /formal argument/, bug10545)
-  end
-
   def test_null_range_cmdarg
     bug10957 = '[ruby-core:68477] [Bug #10957]'
     assert_ruby_status(['-c', '-e', 'p ()..0'], "", bug10957)
@@ -808,6 +819,46 @@ eom
     end
   end
 
+  def test_warning_literal_in_condition
+    assert_warn(/literal in condition/) do
+      eval('1 if ""')
+    end
+    assert_warn(/literal in condition/) do
+      eval('1 if //')
+    end
+    assert_warn(/literal in condition/) do
+      eval('1 if true..false')
+    end
+    assert_warning(/literal in condition/) do
+      eval('1 if 1')
+    end
+    assert_warning(/literal in condition/) do
+      eval('1 if :foo')
+    end
+    assert_warning(/literal in condition/) do
+      eval('1 if :"#{"foo".upcase}"')
+    end
+
+    assert_warn('') do
+      eval('1 if !""')
+    end
+    assert_warn('') do
+      eval('1 if !//')
+    end
+    assert_warn('') do
+      eval('1 if !(true..false)')
+    end
+    assert_warning('') do
+      eval('1 if !1')
+    end
+    assert_warning('') do
+      eval('1 if !:foo')
+    end
+    assert_warning('') do
+      eval('1 if !:"#{"foo".upcase}"')
+    end
+  end
+
   def test_alias_symbol
     bug8851 = '[ruby-dev:47681] [Bug #8851]'
     formats = ['%s', ":'%s'", ':"%s"', '%%s(%s)']
@@ -832,6 +883,102 @@ eom
     end
   end
 
+  def test_parenthesised_statement_argument
+    assert_syntax_error("foo(bar rescue nil)", /unexpected modifier_rescue/)
+    assert_valid_syntax("foo (bar rescue nil)")
+  end
+
+  def test_cmdarg_in_paren
+    bug11873 = '[ruby-core:72482] [Bug #11873]'
+    assert_valid_syntax %q{a b{c d}, :e do end}, bug11873
+    assert_valid_syntax %q{a b(c d), :e do end}, bug11873
+    assert_valid_syntax %q{a b{c(d)}, :e do end}, bug11873
+    assert_valid_syntax %q{a b(c(d)), :e do end}, bug11873
+    assert_valid_syntax %q{a b{c d}, 1 do end}, bug11873
+    assert_valid_syntax %q{a b(c d), 1 do end}, bug11873
+    assert_valid_syntax %q{a b{c(d)}, 1 do end}, bug11873
+    assert_valid_syntax %q{a b(c(d)), 1 do end}, bug11873
+  end
+
+  def test_block_after_cmdarg_in_paren
+    bug11873 = '[ruby-core:72482] [Bug #11873]'
+    def bug11873.p(*);end;
+
+    assert_raise(LocalJumpError, bug11873) do
+      bug11873.instance_eval do
+        p p{p p;p(p)}, tap do
+          raise SyntaxError, "should not be passed to tap"
+        end
+      end
+    end
+
+    assert_raise(LocalJumpError, bug11873) do
+      bug11873.instance_eval do
+        p p{p(p);p p}, tap do
+          raise SyntaxError, "should not be passed to tap"
+        end
+      end
+    end
+  end
+
+  def test_do_block_in_hash_brace
+    bug13073 = '[ruby-core:78837] [Bug #13073]'
+    assert_valid_syntax 'p :foo, {a: proc do end, b: proc do end}', bug13073
+    assert_valid_syntax 'p :foo, {:a => proc do end, b: proc do end}', bug13073
+    assert_valid_syntax 'p :foo, {"a": proc do end, b: proc do end}', bug13073
+    assert_valid_syntax 'p :foo, {** proc do end, b: proc do end}', bug13073
+    assert_valid_syntax 'p :foo, {proc do end => proc do end, b: proc do end}', bug13073
+  end
+
+  def test_do_after_local_variable
+    obj = Object.new
+    def obj.m; yield; end
+    result = assert_nothing_raised(SyntaxError) do
+      obj.instance_eval("m = 1; m do :ok end")
+    end
+    assert_equal(:ok, result)
+  end
+
+  def test_return_toplevel
+    feature4840 = '[ruby-core:36785] [Feature #4840]'
+    code = "#{<<~"begin;"}\n#{<<~"end;"}"
+    begin;
+      return; raise
+      begin return; rescue SystemExit; exit false; end
+      begin return; ensure exit false; end
+      begin ensure return; end
+      begin raise; ensure; return; end
+      begin raise; rescue; return; end
+      return false; raise
+      return 1; raise
+    end;
+    all_assertions(feature4840) do |a|
+      code.each_line do |s|
+        s.chomp!
+        a.for(s) do
+          assert_ruby_status([], s, proc {RubyVM::InstructionSequence.compile(s).disasm})
+        end
+      end
+    end
+  end
+
+  def test_syntax_error_in_rescue
+    bug12613 = '[ruby-core:76531] [Bug #12613]'
+    assert_syntax_error("#{<<-"begin;"}\n#{<<-"end;"}", /Invalid retry/, bug12613)
+    begin;
+      while true
+        begin
+          p
+        rescue
+          retry
+        else
+          retry
+        end
+        break
+      end
+    end;
+  end
+
   private
 
   def not_label(x) @result = x; @not_label ||= nil end
