@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 require 'rubygems/test_case'
 require 'rubygems/commands/open_command'
 
@@ -9,24 +10,47 @@ class TestGemCommandsOpenCommand < Gem::TestCase
     @cmd = Gem::Commands::OpenCommand.new
   end
 
-  def gem name
+  def gem(name, version = "1.0")
     spec = quick_gem name do |gem|
       gem.files = %W[lib/#{name}.rb Rakefile]
+      gem.version = version
     end
     write_file File.join(*%W[gems #{spec.full_name} lib #{name}.rb])
     write_file File.join(*%W[gems #{spec.full_name} Rakefile])
+    spec
   end
 
   def test_execute
     @cmd.options[:args] = %w[foo]
     @cmd.options[:editor] = "#{Gem.ruby} -e0 --"
 
-    gem 'foo'
+    spec = gem 'foo'
+    mock = MiniTest::Mock.new
+    mock.expect(:call, true, [spec.full_gem_path])
 
-    use_ui @ui do
-      @cmd.execute
+    Dir.stub(:chdir, mock) do
+      use_ui @ui do
+        @cmd.execute
+      end
     end
 
+    assert mock.verify
+    assert_equal "", @ui.error
+  end
+
+  def test_wrong_version
+    @cmd.options[:version] = "4.0"
+    @cmd.options[:args] = %w[foo]
+
+    gem "foo", "5.0"
+
+    assert_raises Gem::MockGemUi::TermError do
+      use_ui @ui do
+        @cmd.execute
+      end
+    end
+
+    assert_match %r|Unable to find gem 'foo'|, @ui.output
     assert_equal "", @ui.error
   end
 
