@@ -83,9 +83,7 @@ rb_class_clear_method_cache(VALUE klass, VALUE arg)
 void
 rb_clear_cache(void)
 {
-    rb_warning("rb_clear_cache() is deprecated.");
-    INC_GLOBAL_METHOD_STATE();
-    INC_GLOBAL_CONSTANT_STATE();
+    ONLY_FOR_INTERNAL_USE("rb_clear_cache()");
 }
 
 void
@@ -441,13 +439,18 @@ make_method_entry_refined(VALUE owner, rb_method_entry_t *me)
     }
     else {
 	struct {
-	    const struct rb_method_entry_struct *orig_me;
+	    struct rb_method_entry_struct *orig_me;
 	    VALUE owner;
 	} refined;
 
 	rb_vm_check_redefinition_opt_method(me, me->owner);
 
-	refined.orig_me = rb_method_entry_clone(me);
+	refined.orig_me =
+	    rb_method_entry_alloc(me->called_id, me->owner,
+				  me->defined_class ?
+				  me->defined_class : owner,
+				  method_definition_addref(me->def));
+	METHOD_ENTRY_FLAGS_COPY(refined.orig_me, me);
 	refined.owner = owner;
 
 	method_definition_set(me, method_definition_create(VM_METHOD_TYPE_REFINED, me->called_id), (void *)&refined);
@@ -544,7 +547,6 @@ rb_method_entry_make(VALUE klass, ID mid, VALUE defined_class, rb_method_visibil
 	if (RTEST(ruby_verbose) &&
 	    type != VM_METHOD_TYPE_UNDEF &&
 	    (old_def->alias_count == 0) &&
-	    !make_refined &&
 	    old_def->type != VM_METHOD_TYPE_UNDEF &&
 	    old_def->type != VM_METHOD_TYPE_ZSUPER &&
 	    old_def->type != VM_METHOD_TYPE_ALIAS) {
@@ -561,9 +563,9 @@ rb_method_entry_make(VALUE klass, ID mid, VALUE defined_class, rb_method_visibil
 	      default:
 		break;
 	    }
-	    if (iseq && !NIL_P(iseq->body->location.path)) {
-		int line = iseq->body->line_info_table ? FIX2INT(rb_iseq_first_lineno(iseq)) : 0;
-		rb_compile_warning(RSTRING_PTR(iseq->body->location.path), line,
+	    if (iseq) {
+		rb_compile_warning(RSTRING_PTR(iseq->body->location.path),
+				   FIX2INT(iseq->body->location.first_lineno),
 				   "previous definition of %"PRIsVALUE" was here",
 				   rb_id2str(old_def->original_id));
 	    }
@@ -613,7 +615,7 @@ rb_method_entry_make(VALUE klass, ID mid, VALUE defined_class, rb_method_visibil
 	    recv_class = rb_ivar_get((klass), attached);	\
 	    hook_id = singleton_##hook;			\
 	}						\
-	rb_funcall2(recv_class, hook_id, 1, &arg);	\
+	rb_funcallv(recv_class, hook_id, 1, &arg);	\
     } while (0)
 
 static void
@@ -736,17 +738,19 @@ method_entry_get_without_cache(VALUE klass, ID id,
 	    ent->mid = id;
 
 	    if (UNDEFINED_METHOD_ENTRY_P(me)) {
-		ent->me = 0;
-		me = 0;
+		me = ent->me = NULL;
 	    }
 	    else {
 		ent->me = me;
 	    }
 	}
 	else if (UNDEFINED_METHOD_ENTRY_P(me)) {
-	    me = 0;
+	    me = NULL;
 	}
     }
+    else if (UNDEFINED_METHOD_ENTRY_P(me)) {
+	me = NULL;
+    }
 
     if (defined_class_ptr)
 	*defined_class_ptr = defined_class;
@@ -800,7 +804,7 @@ prepare_callable_method_entry(VALUE defined_class, ID id, const rb_method_entry_
     const rb_callable_method_entry_t *cme;
 
     if (me && me->defined_class == 0) {
-	VM_ASSERT(RB_TYPE_P(defined_class, T_ICLASS));
+	VM_ASSERT(RB_TYPE_P(defined_class, T_ICLASS) || RB_TYPE_P(defined_class, T_MODULE));
 	VM_ASSERT(me->defined_class == 0);
 
 	if ((mtbl = RCLASS_CALLABLE_M_TBL(defined_class)) == NULL) {
@@ -819,9 +823,9 @@ prepare_callable_method_entry(VALUE defined_class, ID id, const rb_method_entry_
     }
     else {
 	cme = (const rb_callable_method_entry_t *)me;
+	VM_ASSERT(callable_method_entry_p(cme));
     }
 
-    VM_ASSERT(callable_method_entry_p(cme));
     return cme;
 }
 
@@ -1029,22 +1033,21 @@ rb_export_method(VALUE klass, ID name, rb_method_visibility_t visi)
 {
     rb_method_entry_t *me;
     VALUE defined_class;
-    VALUE origin_class = RCLASS_ORIGIN(klass);
 
-    me = search_method(origin_class, name, &defined_class);
+    me = search_method(klass, name, &defined_class);
     if (!me && RB_TYPE_P(klass, T_MODULE)) {
 	me = search_method(rb_cObject, name, &defined_class);
     }
 
     if (UNDEFINED_METHOD_ENTRY_P(me) ||
 	UNDEFINED_REFINED_METHOD_P(me->def)) {
-	rb_print_undef(klass, name, 0);
+	rb_print_undef(klass, name, METHOD_VISI_UNDEF);
     }
 
     if (METHOD_ENTRY_VISI(me) != visi) {
 	rb_vm_check_redefinition_opt_method(me, klass);
 
-	if (klass == defined_class || origin_class == defined_class) {
+	if (klass == defined_class || RCLASS_ORIGIN(klass) == defined_class) {
 	    METHOD_ENTRY_VISI_SET(me, visi);
 
 	    if (me->def->type == VM_METHOD_TYPE_REFINED && me->def->body.refined.orig_me) {
@@ -1543,7 +1546,7 @@ rb_alias(VALUE klass, ID alias_name, ID original_name)
 	if ((!RB_TYPE_P(klass, T_MODULE)) ||
 	    (orig_me = search_method(rb_cObject, original_name, &defined_class),
 	     UNDEFINED_METHOD_ENTRY_P(orig_me))) {
-	    rb_print_undef(klass, original_name, 0);
+	    rb_print_undef(klass, original_name, METHOD_VISI_UNDEF);
 	}
     }
 
@@ -1557,10 +1560,10 @@ rb_alias(VALUE klass, ID alias_name, ID original_name)
     if (visi == METHOD_VISI_UNDEF) visi = METHOD_ENTRY_VISI(orig_me);
 
     if (orig_me->defined_class == 0) {
-	rb_method_entry_t *alias_me;
-
-	alias_me = rb_add_method(target_klass, alias_name, VM_METHOD_TYPE_ALIAS, (void *)rb_method_entry_clone(orig_me), visi);
-	alias_me->def->original_id = orig_me->called_id;
+	rb_method_entry_make(target_klass, alias_name, target_klass, visi,
+			     VM_METHOD_TYPE_ALIAS, NULL, orig_me->called_id,
+			     (void *)rb_method_entry_clone(orig_me));
+	method_added(target_klass, alias_name);
     }
     else {
 	rb_method_entry_t *alias_me;
@@ -1666,6 +1669,15 @@ rb_mod_public(int argc, VALUE *argv, VALUE module)
  *  defined methods to protected. With arguments, sets the named methods
  *  to have protected visibility.
  *  String arguments are converted to symbols.
+ *
+ *  If a method has protected visibility, it is callable only where
+ *  <code>self</code> of the context is the same as the method.
+ *  (method definition or instance_eval). This behavior is different from
+ *  Java's protected method. Usually <code>private</code> should be used.
+ *
+ *  Note that a protected method is slow because it can't use inline cache.
+ *
+ *  To show a private method on RDoc, use <code>:doc:</code> instead of this.
  */
 
 static VALUE
@@ -1693,6 +1705,8 @@ rb_mod_protected(int argc, VALUE *argv, VALUE module)
  *       private :a
  *     end
  *     Mod.private_instance_methods   #=> [:a, :c]
+ *
+ *  Note that to show a private method on RDoc, use <code>:doc:</code>.
  */
 
 static VALUE
@@ -1847,7 +1861,7 @@ rb_mod_modfunc(int argc, VALUE *argv, VALUE module)
 		me = search_method(rb_cObject, id, 0);
 	    }
 	    if (UNDEFINED_METHOD_ENTRY_P(me)) {
-		rb_print_undef(module, id, 0);
+		rb_print_undef(module, id, METHOD_VISI_UNDEF);
 	    }
 	    if (me->def->type != VM_METHOD_TYPE_ZSUPER) {
 		break; /* normal case: need not to follow 'super' link */
@@ -1864,7 +1878,9 @@ rb_mod_modfunc(int argc, VALUE *argv, VALUE module)
 int
 rb_method_basic_definition_p(VALUE klass, ID id)
 {
-    const rb_method_entry_t *me = rb_method_entry(klass, id);
+    const rb_method_entry_t *me;
+    if (!klass) return TRUE; /* hidden object cannot be overridden */
+    me = rb_method_entry(klass, id);
     return (me && METHOD_ENTRY_BASIC(me)) ? TRUE : FALSE;
 }
 
@@ -1874,9 +1890,9 @@ call_method_entry(rb_thread_t *th, VALUE defined_class, VALUE obj, ID id,
 {
     const rb_callable_method_entry_t *cme =
 	prepare_callable_method_entry(defined_class, id, me);
-    const rb_block_t *passed_block = th->passed_block;
+    VALUE passed_block_handler = vm_passed_block_handler(th);
     VALUE result = vm_call0(th, obj, id, argc, argv, cme);
-    th->passed_block = passed_block;
+    vm_passed_block_handler_set(th, passed_block_handler);
     return result;
 }
 
@@ -1921,7 +1937,7 @@ vm_respond_to(rb_thread_t *th, VALUE klass, VALUE obj, ID id, int priv)
     const rb_method_entry_t *const me =
 	method_entry_get(klass, resid, &defined_class);
 
-    if (!me) return TRUE;
+    if (!me) return -1;
     if (METHOD_ENTRY_BASIC(me)) {
 	return -1;
     }
@@ -1944,8 +1960,8 @@ vm_respond_to(rb_thread_t *th, VALUE klass, VALUE obj, ID id, int priv)
 	    }
 	    else if (!NIL_P(ruby_verbose)) {
 		VALUE location = rb_method_entry_location(me);
-		rb_warn("%"PRIsVALUE"%c""respond_to?(:%"PRIsVALUE") is"
-			" old fashion which takes only one parameter",
+		rb_warn("%"PRIsVALUE"%c""respond_to?(:%"PRIsVALUE") uses"
+			" the deprecated method signature, which takes one parameter",
 			(FL_TEST(klass, FL_SINGLETON) ? obj : klass),
 			(FL_TEST(klass, FL_SINGLETON) ? '.' : '#'),
 			QUOTE_ID(id));
